# Neurodamus
# Copyright 2018 - Blue Brain Project, EPFL
from __future__ import annotations

import gc
import glob
import itertools
import logging
import math
import os
import shutil
from collections import defaultdict
from contextlib import contextmanager
from os import path as ospath
from pathlib import Path

import libsonata

# Internal Plugins
from . import ngv as _ngv
from .cell_distributor import (
    CellDistributor,
    GlobalCellManager,
    LoadBalance,
    LoadBalanceMode,
    VirtualCellPopulation,
)
from .connection_manager import SynapseRuleManager, edge_node_pop_names
from .core import (
    MPI,
    NeuronWrapper as Nd,
    SimulationProgress,
    mpi_no_errors,
    return_neuron_timings,
    run_only_rank0,
)
from .core._engine import EngineBase
from .core._shmutils import SHMUtil
from .core.configuration import (
    CircuitConfig,
    ConfigurationError,
    Feature,
    GlobalConfig,
    SimConfig,
    _SimConfig,
    get_debug_cell_gids,
    make_circuit_config,
)
from .core.coreneuron_configuration import (
    CompartmentMapping,
    CoreConfig,
)
from .core.nodeset import PopulationNodes
from .gap_junction import GapJunctionManager
from .io.sonata_config import ConnectionTypes
from .modification_manager import ModificationManager
from .neuromodulation_manager import NeuroModulationManager
from .replay import MissingSpikesPopulationError, SpikeManager
from .report import create_report
from .report_parameters import (
    CompartmentType,
    ReportType,
    SectionType,
    check_report_parameters,
    create_report_parameters,
)
from .stimulus_manager import StimulusManager
from .target_manager import TargetManager, TargetSpec
from .utils.logging import log_stage, log_verbose
from .utils.memory import DryRunStats, free_event_queues, pool_shrink, print_mem_usage, trim_memory
from .utils.pyutils import cache_errors
from .utils.timeit import TimerManager, timeit
from neurodamus.core.coreneuron_report_config import CoreReportConfig, CoreReportConfigEntry
from neurodamus.core.coreneuron_simulation_config import CoreSimulationConfig
from neurodamus.utils.pyutils import CumulativeError, rmtree


class METypeEngine(EngineBase):
    CellManagerCls = CellDistributor
    InnerConnectivityCls = SynapseRuleManager
    ConnectionTypes = {
        None: SynapseRuleManager,
        ConnectionTypes.Synaptic: SynapseRuleManager,
        ConnectionTypes.GapJunction: GapJunctionManager,
        ConnectionTypes.NeuroModulation: NeuroModulationManager,
    }
    CircuitPrecedence = 0


class CircuitManager:
    """Holds and manages populations and associated nodes and edges

    All nodes must have a name or read from sonata pop name
    As so, Sonata is preferred when using multiple node files
    """

    def __init__(self):
        self.node_managers = {}  # dict {pop_name -> cell_manager}  # nrn pop is None
        self.virtual_node_managers = {}  # same, but for virtual ones (no cells)
        # dict {(src_pop, dst_pop) -> list[synapse_manager]}
        self.edge_managers = defaultdict(list)
        self.alias = {}  # dict {name -> pop_name}
        self.global_manager = GlobalCellManager()
        self.global_target = TargetManager.create_global_target()

    def initialized(self):
        return bool(self.node_managers)

    def register_node_manager(self, cell_manager):
        pop = cell_manager.population_name
        if pop in self.node_managers:
            raise ConfigurationError(f"Already existing node manager for population {pop}")
        self.node_managers[pop] = cell_manager
        self.alias[cell_manager.circuit_name] = pop
        self.global_manager.register_manager(cell_manager)
        if cell_manager.is_initialized():
            self.global_target.append_nodeset(cell_manager.local_nodes)

    def _new_virtual_node_manager(self, circuit):
        """Instantiate a new virtual node manager explicitly."""
        storage = libsonata.NodeStorage(circuit.CellLibraryFile)
        pop_name, _ = circuit.CircuitTarget.split(":")  # Sonata config fills population
        node_size = storage.open_population(pop_name).size
        gid_vec = list(range(1, node_size + 1))
        virtual_cell_manager = VirtualCellPopulation(pop_name, gid_vec)
        self.virtual_node_managers[pop_name] = virtual_cell_manager
        self.global_target.append_nodeset(virtual_cell_manager.local_nodes)
        return virtual_cell_manager

    @staticmethod
    def new_node_manager_bare(circuit: CircuitConfig, target_manager, run_conf, **kwargs):
        engine = circuit.Engine or METypeEngine
        CellManagerCls = engine.CellManagerCls or CellDistributor
        return CellManagerCls(circuit, target_manager, run_conf, **kwargs)

    def new_node_manager(self, circuit, target_manager, run_conf, *, load_balancer=None, **kwargs):
        if circuit.get("PopulationType") == "virtual":
            return self._new_virtual_node_manager(circuit)
        cell_manager = self.new_node_manager_bare(circuit, target_manager, run_conf, **kwargs)
        cell_manager.load_nodes(load_balancer, **kwargs)
        self.register_node_manager(cell_manager)
        return cell_manager

    def get_node_manager(self, name):
        name = self.alias.get(name, name)
        return self.node_managers.get(name)

    def has_population(self, pop_name):
        return pop_name in self.node_managers

    def unalias_pop_keys(self, source, destination):
        """Un-alias population names"""
        return self.alias.get(source, source), self.alias.get(destination, destination)

    def get_edge_managers(self, source, destination):
        edge_pop_keys = self.unalias_pop_keys(source, destination)
        return self.edge_managers.get(edge_pop_keys) or []

    def get_edge_manager(self, source, destination, conn_type=SynapseRuleManager):
        managers = [
            manager
            for manager in self.get_edge_managers(source, destination)
            if isinstance(manager, conn_type)
        ]
        return managers[0] if managers else None

    def get_create_edge_manager(
        self, conn_type, source, destination, src_target, manager_args=(), **kw
    ):
        source, destination = self.unalias_pop_keys(source, destination)
        manager = self.get_edge_manager(source, destination, conn_type)
        if manager:
            return manager

        if not self.has_population(destination):
            raise ConfigurationError("Can't find projection Node population: " + destination)

        src_manager = self.node_managers.get(source) or self.virtual_node_managers.get(source)
        if src_manager is None:  # src manager may not exist -> virtual
            log_verbose("No known population %s. Creating Virtual src for projection", source)
            if conn_type not in {SynapseRuleManager, _ngv.GlioVascularManager}:
                raise ConfigurationError("Custom connections require instantiated source nodes")
            src_manager = VirtualCellPopulation(source, None, src_target.name)

        target_cell_manager = kw["cell_manager"] = self.node_managers[destination]
        kw["src_cell_manager"] = src_manager
        manager = conn_type(*manager_args, **kw)
        self.edge_managers[source, destination].append(manager)
        target_cell_manager.register_connection_manager(manager)
        return manager

    def all_node_managers(self):
        return self.node_managers.values()

    def all_synapse_managers(self):
        return itertools.chain.from_iterable(self.edge_managers.values())

    @staticmethod
    @run_only_rank0
    def write_population_offsets(pop_offsets, alias_pop, virtual_pop_offsets):
        """Write population_offsets where appropriate

        It is needed for retrieving population offsets for reporting and replay at restore time.

        Format population name::gid offset::population alias
        The virtual population offset is also written for synapse replay in restore.
        The data comes from outside because pop_offsets are not initialized
        in a restore scenario.
        """
        # populations_offset is necessary in output_path
        output_path = SimConfig.populations_offset_output_path(create=True)

        with open(output_path, "w", encoding="utf-8") as f:
            f.writelines(
                "{}::{}::{}\n".format(pop or " ", pop_offsets[pop], alias or " ")
                for alias, pop in alias_pop.items()
            )
            f.writelines(
                "{}::{}::{}\n".format(pop, offset, "virtual")
                for pop, offset in virtual_pop_offsets.items()
            )

        # Add a file in save_path too if required
        if SimConfig.save:
            save_path = SimConfig.populations_offset_save_path(create=True)
            shutil.copy(output_path, save_path)

    def get_population_offsets(self):
        pop_offsets = {
            pop_name: node_manager.local_nodes.offset
            for pop_name, node_manager in self.node_managers.items()
        }
        alias_pop = dict(self.alias)
        return pop_offsets, alias_pop

    def get_virtual_population_offsets(self):
        pop_offsets = {
            pop_name: node_manager.local_nodes.offset
            for pop_name, node_manager in self.virtual_node_managers.items()
        }
        return pop_offsets

    @classmethod
    def read_population_offsets(cls, file_path=None):
        """Read population offsets from populations_offset.dat"""
        pop_offsets = {}
        alias_pop = {}
        virtual_pop_offsets = {}
        with open(file_path or SimConfig.populations_offset_restore_path(), encoding="utf-8") as f:
            for line in f:
                pop, offset, alias = line.strip().split("::")
                pop = pop or None
                alias = alias or None
                if alias == "virtual":
                    virtual_pop_offsets[pop] = int(offset)
                else:
                    pop_offsets[pop] = int(offset)
                    alias_pop[alias] = pop

        return pop_offsets, alias_pop, virtual_pop_offsets

    def __del__(self):
        """De-init. Edge managers must be destructed first"""
        del self.edge_managers
        del self.virtual_node_managers
        del self.node_managers


class Node:
    """The Node class is the main entity for a distributed Neurodamus execution.

    Note that this concept of a "Node" differs from both an MPI node, which
    refers to a process in a parallel computing environment, and a node in the
    circuit graph, which represents an individual element or component within
    the simulation's neural network.

    It serves as the orchestrator for the entire simulation, managing the
    parallel execution of the model and distributing the cells across different
    computational ranks. As the primary control structure, Node is responsible
    for coordinating various components involved in the simulation.

    Internally, the Node class instantiates and manages parallel structures,
    dividing the simulation workload among multiple ranks. With the introduction
    of the concept of multiple populations (also known as multi-circuit), the
    Node class takes partial responsibility for handling this logic, aided by
    the :class:`neurodamus.node.CircuitManager` class (accessible via the
    `circuits` property), which manages the different node and edge managers.

    While many lower-level details of the Node's functionality are encapsulated
    within dedicated helper classes, the Node class still exposes an API that
    allows advanced users to control and inspect almost every major step of the
    simulation. For a standard run, users are encouraged to use the higher-level
    `Neurodamus` class instead, which simplifies some of the complexities
    handled by Node.

    The Node class exposes the following public properties:

    - `circuits`: is a :class:`neurodamus.node.CircuitManager` object,
      responsible for managing multiple node and edge managers within the
      simulation.
    - `target_manager`: is a :class:`neurodamus.target_manager.TargetManager`
      object, responsible for managing the targets in the simulation.
    - `stimulus_manager`: is a
      :class:`neurodamus.stimulus_manager.StimulusManager` object, responsible
      for interpreting and instantiating stimulus events.
    - `elec_manager`: The electrode manager, which controls the interaction with
      simulation electrodes.
    - `reports`: A list of Neurodamus Report `hoc` objects, used to generate
      simulation reports.

    Note that, while the Node object owns and manages most of the top-level
    objects in the simulation, the management of cell and synapse objects has
    been delegated to the `Circuits` class, as these are now handled at a lower
    level.

    Technical note:

    - The properties exposed by Node are read-only, with most internal
      attributes being prefixed with an underscore (`_`). Notable internal
      attributes include:

      `self._sonata_circuits`: The SONATA circuits used by the Node
      each represents a node population.

    These details make the Node class versatile and powerful for advanced users
    who need more granular control over the simulation process.
    """

    _default_population = "All"
    """The default population name for e.g. Reports."""

    def __init__(self, config_file, options: dict | None = None):
        """Creates a neurodamus executor
        Args:
            config_file: A Sonata config file
            options: A dictionary of run options typically coming from cmd line
        """
        options = options or {}
        assert isinstance(config_file, str), "`config_file` should be a string"
        assert config_file, "`config_file` cannot be empty"

        if config_file.endswith("BlueConfig"):
            raise ConfigurationError(
                "Legacy format BlueConfig is not supported, please migrate to SONATA config"
            )
        import libsonata

        conf = libsonata.SimulationConfig.from_file(config_file)
        Nd.init(log_filename=conf.output.log_file, log_use_color=options.pop("use_color", True))

        # This is global initialization, happening once, regardless of number of
        # cycles
        log_stage("Setting up Neurodamus configuration")
        self._pc = Nd.pc
        self._spike_vecs = []
        self._spike_populations = []
        Nd.execute("cvode = new CVode()")
        SimConfig.init(config_file, options)

        if SimConfig.use_coreneuron:
            # Instantiate the CoreNEURON artificial cell object which is used to fill up
            # the empty ranks. This need to be done before the circuit is
            # finitialized
            CoreConfig.instantiate_artificial_cell()

        self._run_conf = SimConfig.run_conf
        self._target_manager = TargetManager(self._run_conf)
        self._target_spec = TargetSpec(self._run_conf.get("CircuitTarget"))
        if SimConfig.use_neuron or SimConfig.coreneuron_direct_mode:
            self._sonatareport_helper = Nd.SonataReportHelper(Nd.dt, True)  # noqa: FBT003
        self._sonata_circuits = SimConfig.sonata_circuits
        self._dump_cell_state_gids = get_debug_cell_gids(options)
        self._core_replay_file = ""
        self._is_ngv_run = any(
            c.Engine.__name__ == "NGVEngine" for c in self._sonata_circuits.values() if c.Engine
        )
        self._initial_rss = 0
        self._cycle_i = 0
        self._n_cycles = 1
        self._shm_enabled = False
        self._dry_run_stats = None

        self._reset()

    def _reset(self):
        """Resets internal state for a new simulation cycle.

        Ensures `_run_conf` is a valid dictionary, initializes core attributes,
        and registers global targets and cell managers.

        Note: remember to call Nd.init(...) before to ensure/load neurodamus mods
        """
        if not self._run_conf or not isinstance(self._run_conf, dict):
            raise ValueError("Invalid `_run_conf`: Must be a dictionary for multi-cycle runs.")

        # Init unconditionally
        self._circuits = CircuitManager()
        self._stim_list = None
        self._report_list = None
        self._stim_manager = None
        self._sim_ready = False
        # flag to mark what we already dumped
        self._last_cell_state_dump_t = None

        self._bbss = Nd.BBSaveState()

        # Register the global target and cell manager
        self._target_manager.register_target(self._circuits.global_target)
        self._target_manager.register_cell_manager(self._circuits.global_manager)

    # public 'read-only' properties - object modification on user responsibility
    circuits = property(lambda self: self._circuits)
    target_manager = property(lambda self: self._target_manager)
    stim_manager = property(lambda self: self._stim_manager)
    stims = property(lambda self: self._stim_list)
    reports = property(lambda self: self._report_list)

    def all_circuits(self):
        yield from self._sonata_circuits.values()

    # -
    def load_targets(self):
        """Initialize targets. Nodesets are loaded on demand."""
        for circuit in self.all_circuits():
            log_verbose("Loading targets for circuit %s", circuit.name or "(default)")
            self._target_manager.load_targets(circuit)

    # -
    @mpi_no_errors
    @timeit(name="Compute LB")
    def compute_load_balance(self):
        """In case the user requested load-balance this function instantiates a
        CellDistributor to split cells and balance those pieces across the available CPUs.
        """
        log_stage("Computing Load Balance")
        circuit = None
        for name, circuit in self._sonata_circuits.items():
            if circuit.get("PopulationType") != "virtual":
                logging.info("Activating experimental LB for Sonata circuit '%s'", name)
                break
        if circuit is None:
            logging.warning(
                "Cannot calculate the load balance because no non-virtual circuit is found"
            )
            return None

        if not circuit.CellLibraryFile:
            logging.info(" => No circuit for Load Balancing. Skipping... ")
            return None

        _ = PopulationNodes.offset_freezer()  # Dont offset while in loadbal

        # Info about the cells to be distributed
        target_spec = TargetSpec(circuit.CircuitTarget)
        target = self.target_manager.get_target(target_spec)

        # Check / set load balance mode
        lb_mode = LoadBalance.select_lb_mode(SimConfig, self._run_conf, target)
        if lb_mode == LoadBalanceMode.RoundRobin:
            return None
        if lb_mode == LoadBalanceMode.Memory:
            logging.info("Load Balancing ENABLED. Mode: Memory")
            return self._memory_mode_load_balancing()

        # Build load balancer as per requested options
        node_path = circuit.CellLibraryFile
        pop = target_spec.population
        load_balancer = LoadBalance(lb_mode, node_path, pop, self._target_manager)

        if load_balancer.valid_load_distribution(target_spec):
            logging.info("Load Balancing done.")
            return load_balancer

        logging.info("Could not reuse load balance data. Doing a Full Load-Balance")
        cell_dist = self._circuits.new_node_manager(circuit, self._target_manager, self._run_conf)
        with load_balancer.generate_load_balance(target_spec, cell_dist):
            # Instantiate the circuit cells and synapses to evaluate complexities
            cell_dist.finalize()
            self._circuits.global_manager.finalize()
            SimConfig.update_connection_blocks(self._circuits.alias)
            target_manager = self._target_manager
            self._create_synapse_manager(SynapseRuleManager, circuit, target_manager)

        # reset since we instantiated with RR distribution
        Nd.t = 0.0  # Reset time
        self.clear_model()

        return load_balancer

    def _memory_mode_load_balancing(self):
        filename = f"allocation_r{MPI.size}_c{SimConfig.modelbuilding_steps}.pkl.gz"

        file_exists = ospath.exists(filename)
        MPI.barrier()

        self._dry_run_stats = DryRunStats()
        if file_exists:
            alloc = self._dry_run_stats.import_allocation_stats(filename, self._cycle_i)
        else:
            logging.warning("Allocation file not found. Generating on-the-fly.")

            compute_cell_memory_usage = not Path(DryRunStats._MEMORY_USAGE_FILENAME).exists()
            if not compute_cell_memory_usage:
                self._dry_run_stats.try_import_cell_memory_usage()
            else:
                logging.warning("Cell memory usage file not found. Computing on-the-fly.")
            for circuit in self._sonata_circuits.values():
                if circuit.get("PopulationType") == "biophysical":
                    cell_distributor = CellDistributor(
                        circuit, self._target_manager, self._run_conf
                    )
                    cell_distributor.load_nodes(
                        None,
                        loader_opts={
                            "load_mode": "load_nodes_metype",
                            "dry_run_stats": self._dry_run_stats,
                        },
                    )
                    if compute_cell_memory_usage:
                        cell_distributor.finalize(dry_run_stats_obj=self._dry_run_stats)
            if compute_cell_memory_usage:
                self._dry_run_stats.collect_all_mpi()
                self._dry_run_stats.export_cell_memory_usage()
                # reset since we instantiated
                Nd.t = 0.0  # Reset time
                self.clear_model()

            alloc, _, _ = self._dry_run_stats.distribute_cells_with_validation(
                MPI.size, SimConfig.modelbuilding_steps
            )
        for pop, ranks in alloc.items():
            for rank, gids in ranks.items():
                logging.debug("Population: %s, Rank: %s, Number of GIDs: %s", pop, rank, len(gids))
        return alloc

    # -
    @mpi_no_errors
    @timeit(name="Cell creation")
    def create_cells(self, load_balance=None):
        """Instantiate and distributes the cells of the network.
        Any targets will be updated to know which cells are local to the cpu.
        """
        if SimConfig.dry_run:
            logging.info("Memory usage after inizialization:")
            print_mem_usage()
            self._dry_run_stats = DryRunStats()
            self._dry_run_stats.try_import_cell_memory_usage()
            loader_opts = {"dry_run_stats": self._dry_run_stats}
        else:
            loader_opts = {}

        loader_opts["cycle_i"] = self._cycle_i

        # Check dynamic attributes required before loading cells
        SimConfig.check_cell_requirements(self.target_manager)

        log_stage("LOADING NODES")
        config = SimConfig.cli_options
        if not load_balance:
            logging.info("Load-balance object not present. Continuing Round-Robin...")

        for name, circuit in self._sonata_circuits.items():
            log_stage("Circuit %s", name)
            if config.restrict_node_populations and name not in config.restrict_node_populations:
                logging.warning("Skipped node population (restrict_node_populations)")
                continue
            self._circuits.new_node_manager(
                circuit,
                self._target_manager,
                self._run_conf,
                load_balancer=load_balance,
                loader_opts=loader_opts,
            )

        lfp_weights_file = self._run_conf.get("LFPWeightsPath")
        if lfp_weights_file:
            if SimConfig.use_coreneuron:
                lfp_manager = self._circuits.global_manager._lfp_manager
                cell_managers = self._circuits.global_manager._cell_managers
                population_list = [
                    manager.population_name
                    for manager in cell_managers
                    if manager.population_name is not None
                ]
                lfp_manager.load_lfp_config(lfp_weights_file, population_list)
            else:
                logging.warning("LFP supported only with CoreNEURON.")

        PopulationNodes.freeze_offsets()  # Dont offset further, could change gids

        # Let the cell managers have any final say in the cell objects
        log_stage("FINALIZING CIRCUIT CELLS")

        for cell_manager in self._circuits.all_node_managers():
            log_stage("Circuit %s", cell_manager.circuit_name or "(default)")
            if SimConfig.dry_run:
                cell_manager.finalize(dry_run_stats_obj=self._dry_run_stats)
            else:
                cell_manager.finalize()

        if SimConfig.dry_run:
            self._dry_run_stats.collect_all_mpi()
            self._dry_run_stats.export_cell_memory_usage()
            self._dry_run_stats.estimate_cell_memory()

        # Final bits after we have all cell managers
        self._circuits.global_manager.finalize()
        SimConfig.update_connection_blocks(self._circuits.alias)

    # -
    @mpi_no_errors
    @timeit(name="Synapse creation")
    def create_synapses(self):
        """Create synapses among the cells, handling connections that appear in the config file"""
        log_stage("LOADING CIRCUIT CONNECTIVITY")
        target_manager = self._target_manager
        manager_kwa = {
            "load_offsets": self._is_ngv_run,
            "dry_run_stats": self._dry_run_stats,
        }

        for circuit in self._sonata_circuits.values():
            Engine = circuit.Engine or METypeEngine
            SynManagerCls = Engine.InnerConnectivityCls
            self._create_synapse_manager(SynManagerCls, circuit, target_manager, **manager_kwa)

        MPI.check_no_errors()
        log_stage("Handling projections...")
        for pname, projection in SimConfig.projections.items():
            self._load_projections(pname, projection, **manager_kwa)

        if SimConfig.dry_run:
            self.syn_total_memory = self._dry_run_stats.collect_display_syn_counts()
            return

        log_stage("Configuring connections...")
        for conn_conf in SimConfig.connections.values():
            self._process_connection_configure(conn_conf)

        logging.info("Done, but waiting for all ranks")

    def _create_synapse_manager(self, ctype, conf, *args, **kwargs):
        """Create a synapse manager for intra-circuit connectivity"""
        log_stage("Circuit %s", conf.name or "(default)")
        if not conf.get("nrnPath"):
            logging.info(" => No connectivity set as internal. See projections")
            return

        if SimConfig.cli_options.restrict_connectivity >= 2:
            logging.warning("Skipped connectivity (restrict_connectivity)")
            return

        c_target = TargetSpec(conf.get("CircuitTarget"))
        if c_target.population is None:
            c_target.population = self._circuits.alias.get(conf.name)

        edge_file, *pop = conf.get("nrnPath").split(":")
        edge_pop = pop[0] if pop else None
        src, dst = edge_node_pop_names(edge_file, edge_pop)

        logging.info("Processing edge file %s, pop: %s", edge_file, edge_pop)

        if src and dst and src != dst:
            raise ConfigurationError("Inner connectivity with different populations")

        dst = self.circuits.alias.get(dst, dst)
        if dst not in SimConfig.cli_options.restrict_node_populations:
            logging.warning("Skipped connectivity (restrict_node_populations)")
            return

        manager = self._circuits.get_create_edge_manager(
            ctype, src, dst, c_target, (conf, *args), **kwargs
        )
        if manager.is_file_open:  # Base connectivity
            manager.create_connections()

    def _process_connection_configure(self, conn_conf):
        source_t = TargetSpec(conn_conf["Source"])
        dest_t = TargetSpec(conn_conf["Destination"])
        source_t.population, dest_t.population = self._circuits.unalias_pop_keys(
            source_t.population, dest_t.population
        )
        src_target = self.target_manager.get_target(source_t)
        dst_target = self.target_manager.get_target(dest_t)
        # Loop over population pairs
        for src_pop in src_target.population_names:
            for dst_pop in dst_target.population_names:
                # Loop over all managers having connections between the populations
                for conn_manager in self._circuits.get_edge_managers(src_pop, dst_pop):
                    logging.debug("Using connection manager: %s", conn_manager)
                    conn_manager.configure_connections(conn_conf)

    @mpi_no_errors
    def _load_projections(self, pname, projection, **kw):
        """Check for Projection blocks"""
        target_manager = self._target_manager
        # None, GapJunctions, NeuroGlial, NeuroModulation...
        ptype = projection.get("Type")
        ptype_cls = EngineBase.connection_types.get(ptype)
        source_t = TargetSpec(projection.get("Source"))
        dest_t = TargetSpec(projection.get("Destination"))

        if SimConfig.cli_options.restrict_connectivity >= 1:
            logging.warning("Skipped projections %s->%s (restrict_connectivity)", source_t, dest_t)
            return

        if not ptype_cls:
            raise RuntimeError(f"No Engine to handle connectivity of type '{ptype}'")

        ppath, *pop_name = projection["Path"].split(":")
        edge_pop_name = pop_name[0] if pop_name else None

        logging.info("Processing Edge file: %s", ppath)

        # Update the target spec with the actual populations
        src_pop, dst_pop = edge_node_pop_names(
            ppath, edge_pop_name, source_t.population, dest_t.population
        )
        source_t.population, dest_t.population = self._circuits.unalias_pop_keys(src_pop, dst_pop)
        src_target = self.target_manager.get_target(source_t)
        dst_target = self.target_manager.get_target(dest_t)

        # If the src_pop is not a known node population, allow creating a Virtual one
        src_populations = src_target.population_names or [source_t.population]

        for src_pop in src_populations:
            for dst_pop in dst_target.population_names:
                logging.info(" * %s (Type: %s, Src: %s, Dst: %s)", pname, ptype, src_pop, dst_pop)
                conn_manager = self._circuits.get_create_edge_manager(
                    ptype_cls,
                    src_pop,
                    dst_pop,
                    source_t,
                    (projection, target_manager),
                    **kw,  # args to ptype_cls if creating
                )
                logging.debug("Using connection manager: %s", conn_manager)
                proj_source = ":".join([ppath, *pop_name])
                conn_manager.open_edge_location(proj_source, projection, src_name=src_pop)
                conn_manager.create_connections(source_t.name, dest_t.name)

    @mpi_no_errors
    @timeit(name="Enable Stimulus")
    def enable_stimulus(self):
        """Iterate over any stimulus defined in the config file given to the simulation
        and instantiate them.
        This passes the raw text in field/value pairs to a StimulusManager object to interpret the
        text and instantiate an actual stimulus object.
        """
        if Feature.Stimulus not in SimConfig.cli_options.restrict_features:
            logging.warning("Skipped Stimulus (restrict_features)")
            return

        log_stage("Stimulus Apply.")

        # for each stimulus defined in the config file, request the StimulusManager to
        # instantiate
        self._stim_manager = StimulusManager(self._target_manager)

        for stim in SimConfig.stimuli:
            if stim.get("Mode") == "Extracellular":
                raise ConfigurationError("input_type extracellular_stimulation is not supported")
            target_spec = TargetSpec(stim.get("Target"))

            stim_name = stim["Name"]
            stim_pattern = stim["Pattern"]
            if stim_pattern == "SynapseReplay":
                continue  # Handled by enable_replay
            logging.info(
                " * [STIM] %s (%s): -> %s",
                stim_name,
                stim_pattern,
                target_spec,
            )
            self._stim_manager.interpret(target_spec, stim)

    # -
    @mpi_no_errors
    def enable_replay(self):
        """Activate replay according to config file. Call before connManager.finalize"""
        if Feature.Replay not in SimConfig.cli_options.restrict_features:
            logging.warning("Skipped Replay (restrict_features)")
            return

        log_stage("Handling Replay")

        if SimConfig.use_coreneuron and bool(self._core_replay_file):
            logging.info(" -> [REPLAY] Reusing stim file from previous cycle")
            return

        for stim in SimConfig.stimuli:
            if stim.get("Pattern") != "SynapseReplay":
                continue
            target = stim["Target"]
            source = stim.get("Source")
            stim_name = stim["Name"]

            #  - delay: Spike replays are suppressed until a certain time
            delay = stim.get("Delay", 0.0)
            logging.info(
                " * [SYN REPLAY] %s -> %s (delay: %d)",
                stim_name,
                target,
                delay,
            )
            self._enable_replay(source, target, stim, delay=delay)

    # -
    def _enable_replay(
        self, source, target, stim_conf, tshift=0.0, delay=0.0, connectivity_type=None
    ):
        ptype_cls = EngineBase.connection_types.get(connectivity_type)
        src_target = self.target_manager.get_target(source)
        dst_target = self.target_manager.get_target(target)

        if SimConfig.restore_coreneuron:
            pop_offsets, alias_pop, _virtual_pop_offsets = CircuitManager.read_population_offsets()

        for src_pop in src_target.population_names:
            try:
                log_verbose("Loading replay spikes for population '%s'", src_pop)
                spike_manager = SpikeManager(stim_conf["SpikeFile"], tshift, src_pop)  # Disposable
            except MissingSpikesPopulationError:
                logging.info("  > No replay for src population: '%s'", src_pop)
                continue

            for dst_pop in dst_target.population_names:
                src_pop_str, dst_pop_str = src_pop or "(base)", dst_pop or "(base)"

                if SimConfig.restore_coreneuron:  # Node and Edges managers not initialized
                    src_pop_offset = (
                        pop_offsets[src_pop]
                        if src_pop in pop_offsets
                        else pop_offsets[alias_pop[src_pop]]
                    )
                else:
                    conn_manager = self._circuits.get_edge_manager(src_pop, dst_pop, ptype_cls)
                    if not conn_manager and SimConfig.cli_options.restrict_connectivity >= 1:
                        continue
                    assert conn_manager, f"Missing edge manager for {src_pop_str} -> {dst_pop_str}"
                    src_pop_offset = conn_manager.src_pop_offset

                logging.info(
                    "=> Population pathway %s -> %s. Source offset: %d",
                    src_pop_str,
                    dst_pop_str,
                    src_pop_offset,
                )
                conn_manager.replay(spike_manager, source, target, delay)

    # -
    @mpi_no_errors
    @timeit(name="Enable Modifications")
    def enable_modifications(self):
        """Iterate over any Modification blocks read from the config file and apply them to the
        network. The steps needed are more complex than NeuronConfigures, so the user should not be
        expected to write the hoc directly, but rather access a library of already available mods.
        """
        # mod_mananger gets destroyed when function returns (not required)
        # mod_manager = Nd.ModificationManager(self._target_manager.hoc)
        log_stage("Enabling modifications...")

        mod_manager = ModificationManager(self._target_manager)
        for name, mod_info in SimConfig.modifications.items():
            target_spec = TargetSpec(mod_info["Target"])
            logging.info(" * [MOD] %s: %s -> %s", name, mod_info["Type"], target_spec)
            mod_manager.interpret(target_spec, mod_info)

    def write_and_get_population_offsets(self) -> tuple[dict, dict, dict]:
        """Retrieve population offsets from the circuit or restore them,
        write the offsets, and return them.

        Returns:
            tuple[dict, dict, dict]:
                - pop_offsets: Mapping of population names to GID offsets.
                - alias_pop: Mapping of population aliases to population names.
                - virtual_pop_offsets: Mapping of virtual population names to offsets.
        """
        if self._circuits.initialized():
            pop_offsets, alias_pop = self._circuits.get_population_offsets()
            virtual_pop_offsets = self._circuits.get_virtual_population_offsets()
        else:
            # restore way
            pop_offsets, alias_pop, virtual_pop_offsets = CircuitManager.read_population_offsets()
        self._circuits.write_population_offsets(
            pop_offsets, alias_pop, virtual_pop_offsets=virtual_pop_offsets
        )
        return pop_offsets, alias_pop, virtual_pop_offsets

    # @mpi_no_errors - not required since theres a call inside before make_comm()
    @timeit(name="Enable Reports")
    def enable_reports(self):  # noqa: C901, PLR0912, PLR0915
        """Iterate over reports defined in the config file and instantiate them."""
        log_stage("Reports Enabling")

        # filter: only the enabled ones
        reports_conf = {name: conf for name, conf in SimConfig.reports.items() if conf["Enabled"]}
        self._report_list = []

        pop_offsets, alias_pop, _virtual_pop_offsets = self.write_and_get_population_offsets()
        pop_offsets_alias = pop_offsets, alias_pop

        if SimConfig.use_coreneuron:
            if SimConfig.restore_coreneuron:
                # we copy it first. We will proceed to modify
                # it in update_report_config later in one go
                Path(CoreConfig.report_config_file_save).parent.mkdir(parents=True, exist_ok=True)
                shutil.copy(
                    CoreConfig.report_config_file_restore, CoreConfig.report_config_file_save
                )
            else:
                core_report_config = CoreReportConfig()

        # necessary for restore: we need to update the various reports tend
        # we can do it in one go later
        substitutions = defaultdict(dict)
        cumulative_error = CumulativeError()
        for rep_name, rep_conf in reports_conf.items():
            cumulative_error.is_error_appended = False
            target_spec = TargetSpec(rep_conf["Target"])
            target = self._target_manager.get_target(target_spec)

            # Build final config. On errors log, stop only after all reports processed
            rep_params = create_report_parameters(
                sim_end=self._run_conf["Duration"],
                nd_t=Nd.t,
                output_root=SimConfig.output_root,
                rep_name=rep_name,
                rep_conf=rep_conf,
                target=target,
                buffer_size=SimConfig.report_buffer_size,
                cumulative_error=cumulative_error,
            )
            if cumulative_error.is_error_appended:
                continue
            check_report_parameters(
                rep_params,
                Nd.dt,
                lfp_active=self._circuits.global_manager._lfp_manager._lfp_file,
                cumulative_error=cumulative_error,
            )
            if cumulative_error.is_error_appended:
                continue

            if SimConfig.restore_coreneuron:
                substitutions[rep_params.name]["end_time"] = rep_params.end
                continue  # we dont even need to initialize reports

            # With coreneuron direct mode, enable fast membrane current calculation
            # for i_membrane
            if (
                SimConfig.coreneuron_direct_mode and "i_membrane" in rep_params.report_on
            ) or rep_params.type == ReportType.LFP:
                Nd.cvode.use_fast_imem(1)

            has_gids = len(self._circuits.global_manager.get_final_gids()) > 0
            if not has_gids:
                self._report_list.append(None)
                continue

            report = create_report(
                params=rep_params,
                use_coreneuron=SimConfig.use_coreneuron,
                cumulative_error=cumulative_error,
            )
            if cumulative_error.is_error_appended:
                continue
            self._set_point_list_in_rep_params(rep_params, cumulative_error=cumulative_error)
            if cumulative_error.is_error_appended:
                continue

            if SimConfig.use_coreneuron:
                core_report_config.add_entry(
                    CoreReportConfigEntry.from_report_params(rep_params=rep_params)
                )

            if not SimConfig.use_coreneuron or rep_params.type == ReportType.SYNAPSE:
                report.setup(
                    rep_params=rep_params,
                    global_manager=self._circuits.global_manager,
                    cumulative_error=cumulative_error,
                )
                if cumulative_error.is_error_appended:
                    continue

            self._report_list.append(report)

        if SimConfig.restore_coreneuron:
            CoreReportConfig.update_file(CoreConfig.report_config_file_save, substitutions)

        cumulative_error.raise_if_any()

        MPI.check_no_errors()

        if not SimConfig.restore_coreneuron:
            if SimConfig.use_coreneuron:
                self._finalize_corenrn_reports(core_report_config, pop_offsets_alias)
            else:
                self._finalize_nrn_reports()

    def _finalize_corenrn_reports(self, core_report_config, pop_offsets_alias):
        core_report_config.set_pop_offsets(pop_offsets_alias[0])
        core_report_config.set_spike_filename(self._run_conf.get("SpikesFile"))
        core_report_config.dump(CoreConfig.report_config_file_save)

    def _finalize_nrn_reports(self):
        # once all reports are created, we finalize the communicator for any reports
        self._sonatareport_helper.set_max_buffer_size_hint(SimConfig.report_buffer_size)
        self._sonatareport_helper.make_comm()
        self._sonatareport_helper.prepare_datasets()

    @cache_errors
    def _set_point_list_in_rep_params(self, rep_params):
        """Dispatcher: it helps to retrieve the points of a target and set them in
        the report parameters.

        Args:
            target: The target name or object
            manager: The cell manager to access gids and metype infos

        Returns: The target list of points
        """
        if rep_params.type == ReportType.COMPARTMENT_SET:
            rep_params.points = rep_params.target.get_point_list_from_compartment_set(
                cell_manager=self._target_manager._cell_manager,
                compartment_set=self._target_manager.get_compartment_set(
                    rep_params.compartment_set
                ),
            )
        else:
            sections, compartments = rep_params.sections, rep_params.compartments
            if rep_params.type == ReportType.SUMMATION and sections == SectionType.SOMA:
                sections, compartments = SectionType.ALL, CompartmentType.ALL
            rep_params.points = rep_params.target.get_point_list(
                cell_manager=self._target_manager._cell_manager,
                section_type=sections,
                compartment_type=compartments,
            )

    # -
    @mpi_no_errors
    def sim_init(self, corenrn_gen=None, **sim_opts):
        """Finalize the model and prepare to run simulation.

        After finalizing the model, will eventually write coreneuron config
        and initialize the neuron simulation if applicable.

        Args:
            corenrn_gen: Whether to generate coreneuron config. Default: None (if required)
            sim_opts - override _finalize_model options. E.g. spike_compress
        """
        if self._sim_ready:
            return self._pc

        if not len(self._circuits.all_node_managers()):
            raise RuntimeError("No CellDistributor was initialized. Please create a circuit.")

        self._finalize_model(**sim_opts)

        if corenrn_gen is None:
            corenrn_gen = SimConfig.use_coreneuron
        if corenrn_gen:
            self._coreneuron_configure_datadir(
                corenrn_restore=False, coreneuron_direct_mode=SimConfig.coreneuron_direct_mode
            )
            self._coreneuron_write_sim_config(corenrn_restore=False)

        if SimConfig.use_neuron or SimConfig.coreneuron_direct_mode:
            self._sim_init_neuron()

        assert not (SimConfig.use_neuron and SimConfig.use_coreneuron)
        if SimConfig.use_neuron:
            self.dump_cell_states()

        self._sim_ready = True
        return self._pc

    # -
    @mpi_no_errors
    @timeit(name="Model Finalized")
    def _finalize_model(self, spike_compress=3):
        """Set up simulation run parameters and initialization.

        Handles setup_transfer, spike_compress, _record_spikes, stdinit, timeout
        Args:
            spike_compress: The spike_compress() parameters (tuple or int)
        """
        logging.info("Preparing to run simulation...")
        for mgr in self._circuits.all_node_managers():
            mgr.pre_stdinit()

        is_save_state = SimConfig.save or SimConfig.restore
        self._pc.setup_transfer()

        if spike_compress and not is_save_state and not self._is_ngv_run:
            # multisend 13 is combination of multisend(1) + two_phase(8) + two_intervals(4)
            # to activate set spike_compress=(0, 0, 13)
            if SimConfig.loadbal_mode == LoadBalanceMode.Memory:
                logging.info("Disabling spike compression for Memory Load Balance")
                spike_compress = False
            if not isinstance(spike_compress, tuple):
                spike_compress = (spike_compress, 1, 0)
            self._pc.spike_compress(*spike_compress)

        # LFP calculation requires WholeCell balancing and extracellular mechanism.
        # This is incompatible with efficient caching atm AND Incompatible with
        # mcd & Glut
        if not self._is_ngv_run:
            Nd.cvode.cache_efficient("ElectrodesPath" not in self._run_conf)
        self._pc.set_maxstep(4)
        with timeit(name="stdinit"):
            Nd.stdinit()

    # -
    def _sim_init_neuron(self):
        # === Neuron specific init ===
        restore_path = SimConfig.restore

        # create a spike_id vector which stores the pairs for spikes and timings for
        # every engine
        for cell_manager in self._circuits.all_node_managers():
            if cell_manager.population_name is not None:
                self._spike_populations.append(
                    (cell_manager.population_name, cell_manager.local_nodes.offset)
                )
                self._spike_vecs.append(cell_manager.record_spikes() or (Nd.Vector(), Nd.Vector()))

        self._pc.timeout(200)  # increase by 10x

        if restore_path:
            with timeit(name="restoretime"):
                logging.info("Restoring state...")
                self._stim_manager.saveStatePreparation(self._bbss)
                self._bbss.vector_play_init()
                self._restart_events()  # On restore the event queue is cleared
                return  # Upon restore sim is ready

    # -
    def _restart_events(self):
        logging.info("Restarting connections events (Replay and Spont Minis)")
        for syn_manager in self._circuits.all_synapse_managers():
            syn_manager.restart_events()

    @contextmanager
    def _coreneuron_ensure_all_ranks_have_gids(self, corenrn_data):
        local_gid_count = sum(
            len(manager.local_nodes) for manager in self._circuits.all_node_managers()
        )
        if local_gid_count > 0:
            yield
            return

        # Create a dummy cell manager with node_pop = None
        # which holds a fake node with a fake population "zzz" to get an unused gid.
        # coreneuron fails if this edge case is reached multiple times as we
        # try to add twice the same gid. pop "zzz" is reserved to be used
        # exclusively for handling cases where no real GIDs are assigned to
        # a rank, ensuring that CoreNeuron does not crash due to missing GIDs.
        log_verbose("Creating fake gid for CoreNeuron")
        assert not PopulationNodes.get("zzz"), "Population 'zzz' is reserved "
        "for handling empty GID ranks and should not be used elsewhere."
        pop_group = PopulationNodes.get("zzz", create=True)
        fake_gid = pop_group.offset + 1 + MPI.rank
        # Add the fake cell to a dummy manager
        dummy_cell_manager = CellDistributor(
            circuit_conf=make_circuit_config({"CellLibraryFile": "<NONE>"}),
            target_manager=self._target_manager,
        )
        dummy_cell_manager.load_artificial_cell(fake_gid, CoreConfig.artificial_cell_object)
        yield

        # register_mapping() doesn't work for this artificial cell as somatic attr is
        # missing, so create a dummy mapping file manually, required for reporting
        cur_files = glob.glob(f"{corenrn_data}/*_3.dat")
        example_mapfile = cur_files[0]
        with open(example_mapfile, "rb") as f_mapfile:
            # read the version from the existing mapping file generated by coreneuron
            coredata_version = f_mapfile.readline().rstrip().decode("ascii")

        mapping_file = Path(corenrn_data, f"{fake_gid}_3.dat")
        if not mapping_file.is_file():
            mapping_file.write_text(f"{coredata_version}\n0\n", encoding="utf-8")

    def _coreneuron_configure_datadir(self, corenrn_restore, coreneuron_direct_mode):
        """Configures the CoreNEURON data directory and handles shared memory (SHM) setup.

        - Creates the data directory if it doesn't exist.
        - If in direct mode, returns immediately since the default behavior is fine.
        - If restoring, skips the setup.
        - If not restoring, checks if SHM should be enabled based on available memory,
          and sets up symlinks for CoreNEURON's necessary files in SHM.

        Args:
            corenrn_restore (bool): Flag indicating if CoreNEURON is in restore mode.
            coreneuron_direct_mode (bool): Flag indicating if direct mode is enabled.
        """
        corenrn_datadir = SimConfig.coreneuron_datadir_path(create=True)
        if coreneuron_direct_mode:
            return
        corenrn_datadir_shm = SHMUtil.get_datadir_shm(corenrn_datadir)

        # Clean-up any previous simulations in the same output directory
        if self._cycle_i == 0 and corenrn_datadir_shm:
            rmtree(corenrn_datadir_shm)

        # Ensure that we have a folder in /dev/shm (i.e., 'SHMDIR' ENV variable)
        if SimConfig.cli_options.enable_shm and not corenrn_datadir_shm:
            logging.warning("Unknown SHM directory for model file transfer in CoreNEURON.")
        # Try to configure the /dev/shm folder as the output directory for the files
        elif (
            self._cycle_i == 0
            and not corenrn_restore
            and (SimConfig.cli_options.enable_shm and SimConfig.delete_corenrn_data)
        ):
            # Check for the available memory in /dev/shm and estimate the RSS by multiplying
            # the number of cycles in the multi-step model build with an approximate
            # factor
            mem_avail = SHMUtil.get_mem_avail()
            shm_avail = SHMUtil.get_shm_avail()
            initial_rss = self._initial_rss
            current_rss = SHMUtil.get_node_rss()
            factor = SHMUtil.get_shm_factor()
            rss_diff = (current_rss - initial_rss) if initial_rss < current_rss else current_rss
            # 'rss_diff' prevents <0 estimates
            rss_req = int(rss_diff * self._n_cycles * factor)

            # Sync condition value with all ranks to ensure that all of them can use
            # /dev/shm
            shm_possible = (rss_req < shm_avail) and (rss_req < mem_avail)
            if MPI.allreduce(int(shm_possible), MPI.SUM) == MPI.size:
                logging.info("SHM file transfer mode for CoreNEURON enabled")

                # Create SHM folder and links to GPFS for the global data structures
                os.makedirs(corenrn_datadir_shm, exist_ok=True)

                # Important: These three files must be available on every node, as they are shared
                #            across all of the processes. The trick here is to fool NEURON into
                #            thinking that the files are written in /dev/shm, but they are actually
                #            written on GPFS. The workflow is identical, meaning that rank 0 writes
                #            the content and every other rank reads it afterwards in CoreNEURON.
                for filename in ("bbcore_mech.dat", "files.dat", "globals.dat"):
                    path = os.path.join(corenrn_datadir, filename)
                    path_shm = os.path.join(corenrn_datadir_shm, filename)

                    try:
                        os.close(os.open(path, os.O_CREAT))
                        os.symlink(path, path_shm)
                    except FileExistsError:
                        pass  # Ignore if other process has already created it

                # Update the flag to confirm the configuration
                self._shm_enabled = True
            else:
                logging.warning(
                    "Unable to utilize SHM for model file transfer in CoreNEURON. "
                    "Increase the number of nodes to reduce the memory footprint "
                    "(Current use node: %d MB / SHM Limit: %d MB / Mem. Limit: %d MB)",
                    (rss_req >> 20),
                    (shm_avail >> 20),
                    (mem_avail >> 20),
                )
        _SimConfig.coreneuron_datadir = (
            corenrn_datadir if not self._shm_enabled else corenrn_datadir_shm
        )

    # -
    @timeit(name="corewrite")
    def _coreneuron_write_sim_config(self, corenrn_restore):
        log_stage("Dataset generation for CoreNEURON")

        if not corenrn_restore:
            CompartmentMapping(self._circuits.global_manager).register_mapping()
            if not SimConfig.coreneuron_direct_mode:
                with self._coreneuron_ensure_all_ranks_have_gids(CoreConfig.datadir):
                    self._pc.nrnbbcore_write(CoreConfig.datadir)
                    MPI.barrier()  # wait for all ranks to finish corenrn data generation

        prcellgid = self._dump_cell_state_gids[0] if self._dump_cell_state_gids else -1
        if self._dump_cell_state_gids and len(self._dump_cell_state_gids) > 1:
            logging.warning(
                "Multiple cell state GIDs provided. Using the first one: %d",
                self._dump_cell_state_gids[0],
            )

        core_simulation_config = CoreSimulationConfig(
            outpath=CoreConfig.output_root,
            datpath=CoreConfig.datadir,
            tstop=Nd.tstop,
            dt=Nd.dt,
            prcellgid=prcellgid,
            celsius=getattr(SimConfig, "celsius", 34.0),
            voltage=getattr(SimConfig, "v_init", -65.0),
            cell_permute=int(SimConfig.cell_permute),
            pattern=self._core_replay_file or None,
            seed=SimConfig.rng_info.getGlobalSeed(),
            model_stats=int(SimConfig.cli_options.model_stats),
            report_conf=CoreConfig.report_config_file_save
            if self._run_conf["EnableReports"]
            else None,
            mpi=int(os.environ.get("NEURON_INIT_MPI", "1")),
        )
        core_simulation_config.dump(CoreConfig.sim_config_file)
        # Wait for rank0 to write the sim config file
        MPI.barrier()
        logging.info(" => Dataset written to '%s'", CoreConfig.datadir)

    # -
    def run_all(self):
        """Run the whole simulation according to the simulation config file"""
        if not self._sim_ready:
            self.sim_init()

        timings = None
        if SimConfig.use_neuron:
            timings = self._run_neuron()
            self.sonata_spikes()
        if SimConfig.use_coreneuron:
            print_mem_usage()
            if not SimConfig.coreneuron_direct_mode:
                self.clear_model(avoid_clearing_queues=False)
            self._run_coreneuron()
            if SimConfig.coreneuron_direct_mode:
                self.sonata_spikes()
        return timings

    # -
    @return_neuron_timings
    def _run_neuron(self):
        if MPI.rank == 0:
            _ = SimulationProgress()
        self.solve()
        logging.info("Simulation finished.")

    @staticmethod
    def _run_coreneuron():
        logging.info("Launching simulation with CoreNEURON")
        CoreConfig.psolve_core(
            SimConfig.coreneuron_direct_mode,
        )

    def _sim_event_handlers(self, tstart, tstop):
        """Create handlers for "in-simulation" events, like activating delayed
        connections, execute Save-State, etc
        """
        events = defaultdict(list)  # each key (time) points to a list of handlers

        if SimConfig.save:
            tsave = SimConfig.tstop  # Consider 0 as the end too!
            save_f = self._create_save_handler()
            events[tsave].append(save_f)

        event_list = [(t, events[t]) for t in sorted(events)]
        return event_list

    # -
    def _create_save_handler(self):
        @timeit(name="savetime")
        def save_f():
            """Function that saves the current simulation state:
            syncs MPI, saves stimuli, flushes reports, clears the model,
            and logs progress.
            """
            logging.info("Saving State... (t=%f)", SimConfig.tstop)
            MPI.barrier()
            self._stim_manager.saveStatePreparation(self._bbss)
            log_verbose("SaveState Initialization Done")

            # If event at the end of the sim we can actually clearModel()
            # before savestate()
            log_verbose("Clearing model prior to final save")
            self._sonatareport_helper.flush()

            self.clear_model()
            logging.info(" => Save done successfully")

        return save_f

    # -
    @mpi_no_errors
    @timeit(name="psolve")
    def solve(self, tstop=None):
        """Call solver with a given stop time (default: whole interval).
        Be sure to have sim_init()'d the simulation beforehand
        """
        if not self._sim_ready:
            raise ConfigurationError("Initialize simulation first")

        tstart = Nd.t
        tstop = tstop or Nd.tstop
        event_list = self._sim_event_handlers(tstart, tstop)

        # NOTE: _psolve_loop is called among events in order to eventually split long
        # simulation blocks, where one or more report flush(es) can happen. It is a simplified
        # design relatively to the original version where the report checkpoint would not happen
        # before the checkpoint timeout (25ms default). However there shouldn't be almost any
        # performance penalty since the simulation is already halted between events.

        logging.info("Running simulation until t=%d ms", tstop)
        t = tstart  # default if there are no events
        for t, events in event_list:
            self._psolve_loop(t)
            for event in events:
                event()
            self.dump_cell_states()
        # Run until the end
        if t < tstop:
            self._psolve_loop(tstop)
            self.dump_cell_states()

        # Final flush
        self._sonatareport_helper.flush()

    # psolve_loop: There was an issue where MPI collective routines for reporting and spike exchange
    # are mixed such that some cpus are blocked waiting to complete reporting while others to
    # finish spike exchange. As a work-around, periodically halt simulation and flush reports
    # Default is 25 ms / cycle
    def _psolve_loop(self, tstop):
        cur_t = round(Nd.t, 2)  # fp innnacuracies could lead to infinitesimal loops
        buffer_t = SimConfig.buffer_time
        for _ in range(math.ceil((tstop - cur_t) / buffer_t)):
            next_flush = min(tstop, cur_t + buffer_t)
            self._pc.psolve(next_flush)
            cur_t = next_flush
        Nd.t = cur_t

    # -
    @mpi_no_errors
    def clear_model(self, avoid_creating_objs=False, avoid_clearing_queues=True):
        """Clears appropriate lists and other stored references.
        For use with intrinsic load balancing. After creating and evaluating the network using
        round robin distribution, we want to clear the cells and synapses in order to have a
        clean slate on which to instantiate the balanced cells.
        """
        logging.info("Clearing model")
        self._pc.gid_clear()
        self._target_manager.clear_simulation_data()

        if not avoid_creating_objs and SimConfig.use_neuron and self._sonatareport_helper:
            self._sonatareport_helper.clear()

        # Reset vars
        self._reset()

        # Clear BBSaveState
        self._bbss.ignore()

        # Shrink ArrayPools holding mechanism's data in NEURON
        pool_shrink()

        # Free event queues in NEURON
        if not avoid_clearing_queues:
            free_event_queues()

        # Garbage collect all Python objects without references
        gc.collect()

        # Finally call malloc_trim to return all the freed pages back to the OS
        trim_memory()
        print_mem_usage()

    # -------------------------------------------------------------------------
    #  output
    # -------------------------------------------------------------------------

    def sonata_spikes(self):
        """Write the spike events that occured on each node into a single output SONATA file."""
        output_root = SimConfig.output_root_path(create=True)
        if hasattr(self._sonatareport_helper, "create_spikefile"):
            # Write spike report for multiple populations if exist
            spike_path = self._run_conf.get("SpikesFile")

            # Get only the spike file name
            file_name = spike_path.split("/")[-1] if spike_path is not None else "out.h5"

            # create a sonata spike file
            self._sonatareport_helper.create_spikefile(output_root, file_name)
            # write spikes per population
            for (population, population_offset), (spikevec, idvec) in zip(
                self._spike_populations, self._spike_vecs
            ):
                extra_args = (
                    (population, population_offset) if population else ("All", population_offset)
                )
                self._sonatareport_helper.add_spikes_population(spikevec, idvec, *extra_args)
            # write all spike populations
            self._sonatareport_helper.write_spike_populations()
            # close the spike file
            self._sonatareport_helper.close_spikefile()
        else:
            # fallback: write spike report with one single population "ALL"
            logging.warning(
                "Writing spike reports with multiple populations is not supported. "
                "If needed, please update to a newer version of neurodamus."
            )
            population = self._target_spec.population or "All"
            extra_args = (population,)
            self._sonatareport_helper.write_spikes(spikevec, idvec, output_root, *extra_args)

    def dump_cell_states(self):
        """Dump the _pr_cell_gid cell state if not already done

        We assume that the parallel context is ready. Thus, This function should
        not be called if coreNeuron is employed and we are not at t=0.0.
        """
        assert SimConfig.use_neuron, "This function can work only with Neuron. Use sim.conf to "
        "instruct coreNeuron to dump a cell state instead."
        if not self._dump_cell_state_gids:
            return
        if self._last_cell_state_dump_t == Nd.t:  # avoid duplicating output
            return

        for i in self._dump_cell_state_gids:
            log_verbose("Dumping info about cell %d", i)

            self._pc.prcellstate(i, f"py_Neuron_t{Nd.t}")

        self._last_cell_state_dump_t = Nd.t

    @staticmethod
    @run_only_rank0
    def coreneuron_cleanup():
        """Clean coreneuron save files after running"""
        data_folder = Path(CoreConfig.datadir)
        logging.info("Deleting intermediate data in %s", data_folder)
        assert data_folder.is_dir(), "Data folder must be a directory"
        if data_folder.is_symlink():
            # in restore, coreneuron data is a symbolic link
            data_folder.unlink()
        else:
            rmtree(data_folder)

        build_path = Path(SimConfig.build_path())
        if build_path.exists():
            shutil.rmtree(build_path)

        sim_conf = Path(CoreConfig.sim_config_file)
        assert not sim_conf.exists()

        report_file = Path(CoreConfig.report_config_file_save)
        assert not report_file.exists()

    def cleanup(self):
        """Have the compute nodes wrap up tasks before exiting."""
        # MemUsage constructor will do MPI communications
        print_mem_usage()

        # Coreneuron runs clear the model before starting
        if not SimConfig.use_coreneuron or SimConfig.simulate_model is False:
            self.clear_model(avoid_creating_objs=True)

        if SimConfig.delete_corenrn_data and not SimConfig.save and not SimConfig.dry_run:
            with timeit(name="Delete corenrn data"):
                self.coreneuron_cleanup()
                MPI.barrier()

    @staticmethod
    @run_only_rank0
    def move_dumpcellstates_to_output_root():
        """Check for .corenrn or .nrn files in the current directory
        and move them to CoreConfig.output_root_path(create=True).
        """
        current_dir = Path(".")
        output_root = Path(SimConfig.output_root_path(create=True))

        # Iterate through files in the current directory
        for file in current_dir.iterdir():
            if file.suffix in {".corenrn", ".nrn", ".nrndat"}:
                shutil.move(str(file), output_root / file.name)
                logging.info("Moved %s to %s", file.name, output_root)


class Neurodamus(Node):
    """A high level interface to Neurodamus"""

    def __init__(self, config_file, auto_init=True, logging_level=None, **user_opts):
        """Creates and initializes a neurodamus run node

        As part of Initiazation it calls:
         * load_targets
         * compute_load_balance
         * Build the circuit (cells, synapses, GJs)
         * Add stimulus & replays
         * Activate reports if requested

        Args:
            config_file: The simulation config recipe file
            logging_level: (int) Redefine the global logging level.
                0 - Only warnings / errors
                1 - Info messages (default)
                2 - Verbose
                3 - Debug messages
            user_opts: Options to Neurodamus overriding the simulation config file
        """
        enable_reports = not user_opts.pop("disable_reports", False)
        if logging_level is not None:
            GlobalConfig.verbosity = logging_level

        Node.__init__(self, config_file, user_opts)
        # Use the run_conf dict to avoid passing it around
        self._run_conf["EnableReports"] = enable_reports
        self._run_conf["AutoInit"] = auto_init

        if SimConfig.dry_run:
            if self._is_ngv_run:
                raise Exception("Dry run not available for ngv circuit")
            self.load_targets()
            self.create_cells()
            self.create_synapses()
            return

        if SimConfig.restore_coreneuron:
            self._coreneuron_restore()
        elif SimConfig.build_model:
            self._instantiate_simulation()

        # Remove .SUCCESS file if exists
        self._success_file = SimConfig.config_file + ".SUCCESS"
        self._remove_file(self._success_file)

    # -
    def _build_single_model(self):
        """Construct the model for a single cycle.

        This process includes:
        - Computing load balance across ranks.
        - Building the circuit by creating cells and applying configurations.
        - Establishing synaptic connections.
        - Enabling replay mechanisms if applicable.
        - Initializing the simulation if 'AutoInit' is enabled.
        """
        log_stage("================ CALCULATING LOAD BALANCE ================")
        load_bal = self.compute_load_balance()
        print_mem_usage()

        log_stage("==================== BUILDING CIRCUIT ====================")
        self.create_cells(load_bal)
        print_mem_usage()

        # Create connections
        self.create_synapses()
        print_mem_usage()

        log_stage("================ INSTANTIATING SIMULATION ================")
        # Apply replay
        self.enable_replay()
        print_mem_usage()

        if self._run_conf["AutoInit"]:
            self.init()

    # -
    def init(self):
        """Explicitly initialize, allowing users to make last changes before simulation"""
        if self._sim_ready:
            logging.warning("Simulation already initialized. Skip second init")
            return

        log_stage("Creating connections in the simulator")
        base_seed = self._run_conf.get("BaseSeed", 0)  # base seed for synapse RNG
        for syn_manager in self._circuits.all_synapse_managers():
            syn_manager.finalize(base_seed)
        print_mem_usage()

        self.enable_stimulus()
        print_mem_usage()
        self.enable_modifications()

        if self._run_conf["EnableReports"]:
            self.enable_reports()
        print_mem_usage()

        self.sim_init()
        assert self._sim_ready, "sim_init should have set this"

    @staticmethod
    def _merge_filesdat(ncycles):
        log_stage("Generating merged CoreNeuron files.dat")
        coreneuron_datadir = CoreConfig.datadir
        cn_entries = []
        for i in range(ncycles):
            log_verbose(f"files_{i}.dat")
            filename = ospath.join(coreneuron_datadir, f"files_{i}.dat")
            with open(filename, encoding="utf-8") as fd:
                first_line = fd.readline()
                nlines = int(fd.readline())
                for _ in range(nlines):
                    line = fd.readline()
                    cn_entries.append(line)

        cnfilename = ospath.join(coreneuron_datadir, "files.dat")
        with open(cnfilename, "w", encoding="utf-8") as cnfile:
            cnfile.write(first_line)
            cnfile.write(str(len(cn_entries)) + "\n")
            cnfile.writelines(cn_entries)

        logging.info(" => %s files merged successfully", ncycles)

    def _coreneuron_restore(self):
        """Restore CoreNEURON simulation state.

        This method sets up the CoreNEURON environment for restoring a simulation:
        - load targets
        - enable replay
        - enable reports (this writes also report.conf)
        - write sim.conf
        - set and link coreneuron_datadir to the old restore one
        """
        log_stage(" =============== CORENEURON RESTORE ===============")
        self.load_targets()
        self.enable_replay()
        if self._run_conf["EnableReports"]:
            self.enable_reports()

        self._coreneuron_write_sim_config(corenrn_restore=True)
        self._setup_coreneuron_datadir_from_restore()

        self._sim_ready = True

    @run_only_rank0
    def _setup_coreneuron_datadir_from_restore(self):
        """Configure the environment for restoring CoreNEURON.

        This involves:
        - setting the coreneuron_datadir
        - writing the sim.conf
        - linking the old coreneuron_datadir to the new one
        (in save_path or output_root)
        """
        self._coreneuron_configure_datadir(
            corenrn_restore=True, coreneuron_direct_mode=SimConfig.coreneuron_direct_mode
        )

        # handle coreneuron_input movements
        src_datadir = Path(SimConfig.coreneuron_datadir_restore_path())
        dst_datadir = Path(SimConfig.coreneuron_datadir_path())
        # Check if source directory exists
        if not src_datadir.exists():
            raise FileNotFoundError(
                f"Coreneuron input directory in `{src_datadir}` does not exist!"
            )

        # If the source exists,
        # remove the destination directory or symlink (if it exists)
        if dst_datadir.exists():
            if dst_datadir.is_symlink():
                # Remove the symlink
                dst_datadir.unlink()
            else:
                # Remove the folder if it's not a symlink
                shutil.rmtree(dst_datadir)

        dst_datadir.symlink_to(src_datadir)

    def compute_n_cycles(self):
        """Determine the number of model-building cycles

        It is based on configuration and system constraints.
        """
        n_cycles = SimConfig.modelbuilding_steps
        # No multi-cycle. Trivial result, this is always possible
        if n_cycles == 1:
            return n_cycles

        target = self._target_manager.get_target(self._target_spec)
        target_name = self._target_spec.name
        max_cell_count = target.max_gid_count_per_population()
        logging.info(
            "Simulation target: %s, Max cell count per population: %d", target_name, max_cell_count
        )

        if SimConfig.use_coreneuron and max_cell_count / n_cycles < MPI.size and max_cell_count > 0:
            # coreneuron with no. ranks >> no. cells
            # need to assign fake gids to artificial cells in empty threads
            # during module building fake gids start from max_gid + 1
            # currently not support engine plugin where target is loaded later
            # We can always have only 1 cycle. coreneuron throws an error if a
            # rank does not have cells during a cycle. There is a way to prevent
            # this for unbalanced multi-populations but if more than one cycle
            # happens on a rank without instantiating cells another error raises.
            # Thus, the number of cycles should be rounded down; on the safe side
            max_num_cycles = int(max_cell_count / MPI.size) or 1
            if n_cycles > max_num_cycles:
                logging.warning(
                    "Your simulation is using multi-cycle without enough cells.\n"
                    "  => Number of cycles has been automatically set to the max: %d",
                    max_num_cycles,
                )
                n_cycles = max_num_cycles
        return n_cycles

    def _build_model(self):
        """Build the model

        Internally it calls _build_single_model, over multiple
        cycles if necessary.

        Note: only relevant for coreNeuron
        """
        self._n_cycles = self.compute_n_cycles()

        # Without multi-cycle, it's a trivial model build.
        # sub_targets is False
        if self._n_cycles == 1:
            self._build_single_model()
            return

        logging.info("MULTI-CYCLE RUN: %s Cycles", self._n_cycles)
        target = self._target_manager.get_target(self._target_spec)
        TimerManager.archive(archive_name="Before Cycle Loop")

        PopulationNodes.freeze_offsets()

        if SimConfig.loadbal_mode != LoadBalanceMode.Memory:
            sub_targets = target.generate_subtargets(self._n_cycles)

        for cycle_i in range(self._n_cycles):
            logging.info("")
            logging.info("-" * 60)
            log_stage(f"==> CYCLE {cycle_i + 1} (OUT OF {self._n_cycles})")
            logging.info("-" * 60)

            self.clear_model()

            if SimConfig.loadbal_mode != LoadBalanceMode.Memory:
                for cur_target in sub_targets[cycle_i]:
                    self._target_manager.register_target(cur_target)
                    pop = next(iter(cur_target.population_names))
                    for circuit in self._sonata_circuits.values():
                        tmp_target_spec = TargetSpec(circuit.CircuitTarget)
                        if tmp_target_spec.population == pop:
                            tmp_target_spec.name = cur_target.name
                            circuit.CircuitTarget = str(tmp_target_spec)

            self._cycle_i = cycle_i
            self._build_single_model()

            # Move generated files aside (to be merged later)
            if MPI.rank == 0:
                base_filesdat = ospath.join(CoreConfig.datadir, "files")
                os.rename(base_filesdat + ".dat", base_filesdat + f"_{cycle_i}.dat")
            # Archive timers for this cycle
            TimerManager.archive(archive_name=f"Cycle Run {cycle_i + 1:d}")

        if MPI.rank == 0:
            self._merge_filesdat(self._n_cycles)

    # -
    def _instantiate_simulation(self):
        """Initialize the simulation

        - load targets
        - check connections
        - build the model
        """
        # Keep the initial RSS for the SHM file transfer calculations
        self._initial_rss = SHMUtil.get_node_rss()
        print_mem_usage()

        self.load_targets()

        # Check connection block configuration and raise warnings for overriding
        # parameters
        SimConfig.check_connections_configure(self._target_manager)

        self._build_model()

    # -
    @timeit(name="finished Run")
    def run(self, cleanup=True):
        """Prepares and launches the simulation according to the loaded config.
        If '--only-build-model' option is set, simulation is skipped.

        Args:
            cleanup (bool): Free up the model and intermediate files [default: true]
                Rationale is: the high-level run() method it's typically for a
                one shot simulation so we should cleanup. If not it can be set to False
        """
        if SimConfig.dry_run:
            log_stage("============= DRY RUN (SKIP SIMULATION) =============")
            self._dry_run_stats.display_total()
            self._dry_run_stats.display_node_suggestions()
            ranks = self._dry_run_stats.get_num_target_ranks(SimConfig.num_target_ranks)
            self._dry_run_stats.collect_all_mpi()
            try:
                self._dry_run_stats.distribute_cells_with_validation(
                    ranks, SimConfig.modelbuilding_steps
                )
            except RuntimeError:
                logging.exception("Dry run failed")
            return

        if not SimConfig.simulate_model:
            self.sim_init()
            log_stage("======== [SKIPPED] SIMULATION (MODEL BUILD ONLY) ========")
        elif not SimConfig.build_model:
            log_stage("============= SIMULATION (SKIP MODEL BUILD) =============")
            # coreneuron needs the report file created
            self._run_coreneuron()
        else:
            log_stage("======================= SIMULATION =======================")
            self.run_all()

        # Create SUCCESS file if the simulation finishes successfully
        self._touch_file(self._success_file)
        logging.info("Finished! Creating .SUCCESS file: '%s'", self._success_file)

        # Save seclamp holding currents for gap junction user corrections
        if (
            gj_target_pop := SimConfig.beta_features.get("gapjunction_target_population")
        ) and SimConfig.beta_features.get("procedure_type") == "find_holding_current":
            gj_manager = self._circuits.get_edge_manager(
                gj_target_pop, gj_target_pop, GapJunctionManager
            )
            gj_manager.save_seclamp()

        self.move_dumpcellstates_to_output_root()

        if cleanup:
            self.cleanup()

    @staticmethod
    @run_only_rank0
    def _remove_file(file_name):
        import contextlib

        with contextlib.suppress(FileNotFoundError):
            os.remove(file_name)

    @staticmethod
    @run_only_rank0
    def _touch_file(file_name):
        with open(file_name, "a", encoding="utf-8"):
            os.utime(file_name, None)



"""Implements several helper modules for building circuits with Neuron.

They can be seen as a High-Level Neuron API, and several examples are found under `examples`.
"""

from ._engine import EngineBase
from ._mpi import MPI, OtherRankError
from ._neurodamus import NeuronWrapper
from ._neuron import MComplexLoadBalancer, Neuron
from ._utils import (
    ProgressBarRank0,
    SimulationProgress,
    mpi_no_errors,
    return_neuron_timings,
    run_only_rank0,
)

__all__ = [
    "MPI",
    "EngineBase",
    "MComplexLoadBalancer",
    "Neuron",
    "NeuronWrapper",
    "OtherRankError",
    "ProgressBarRank0",
    "SimulationProgress",
    "mpi_no_errors",
    "return_neuron_timings",
    "run_only_rank0",
]



"""Collection of Cell Readers from different sources (Pure HDF5, SynTool...)"""

import logging
from collections import defaultdict
from os import path as ospath

import libsonata
import numpy as np

from neurodamus.core import NeuronWrapper as Nd, run_only_rank0
from neurodamus.core.configuration import SimConfig
from neurodamus.metype import METypeManager
from neurodamus.utils.logging import log_verbose

EMPTY_GIDVEC = np.empty(0, dtype="uint32")


def split_round_robin(all_gids, stride=1, stride_offset=0, total_cells=None):
    """Splits a numpy ndarray[uint32] round-robin.
    If the array is None generates new arrays based on the nr of total cells
    """
    if all_gids is not None:
        gidvec = all_gids[stride_offset::stride] if stride > 1 else all_gids
        gidvec.sort()
    else:
        assert total_cells, "split_round_robin: total_cells required without gids"
        cell_i = stride_offset + 1  # gids start from 1
        gidvec = np.arange(cell_i, total_cells + 1, stride, dtype="uint32")
    return gidvec


def dry_run_distribution(gid_metype_bundle, stride=1, stride_offset=0):
    """Distribute gid in metype bundles for dry run.

    The principle is the following: all gids with the same metype
    have to be assigned to the same rank. This function receives
    a list of list of gids, each sublist containing gids of the same
    metype. The gid_metype_bundle list of lists is generated by the
    retrieve_unique_metype function. This function performs a
    round robin distribution of the inner lists, i.e. it returns
    a list of gids that are sequentially in the same metype.
    The return is a flattened numpy array of gids that shall be
    instantiated on the same rank.

    Example:
        gid_metype_bundle = [[1, 2, 3], [4, 5, 6], [7, 8, 9], [10]]

        stride = 2
        stride_offset = 0
        return = [1, 2, 3, 7, 8, 9]

        stride = 2
        stride_offset = 1
        return = [4, 5, 6, 10]

    Args:
        gid_metype_bundle: list of lists of gids to be distributed
        mpi_size: MPI size
        mpi_rank: MPI rank
    Returns:
        A numpy array of gids that are sequentially in the same metype
    """
    if not gid_metype_bundle:
        return EMPTY_GIDVEC

    # if mpi_size is 1, return all gids flattened
    if stride == 1:
        return np.concatenate(gid_metype_bundle)
    groups = gid_metype_bundle[stride_offset::stride]
    return np.concatenate(groups) if groups else EMPTY_GIDVEC


def load_sonata(  # noqa: C901, PLR0915
    circuit_conf,
    all_gids,
    stride=1,
    stride_offset=0,
    *,
    node_population,
    load_dynamic_props=(),
    has_extra_data=False,
    dry_run_stats=None,
    load_mode=None,
):
    """A reader supporting additional dynamic properties from Sonata files."""
    import libsonata

    node_file = circuit_conf.CellLibraryFile
    node_store = libsonata.NodeStorage(node_file)
    node_pop = node_store.open_population(node_population)
    attr_names = node_pop.attribute_names
    dynamics_attr_names = node_pop.dynamics_attribute_names
    total_cells = node_pop.size

    def load_base_info_dry_run():
        CELL_NODE_INFO_LIMIT = 100
        log_verbose("Sonata dry run mode: looking for unique metype instances")
        meinfos = METypeManager()
        # Get global METype counts (computed in rank0, broadcasted)
        metype_gids, counts = _retrieve_unique_metypes(node_pop, all_gids)
        dry_run_stats.metype_counts += counts
        dry_run_stats.pop_metype_gids[node_population] = metype_gids
        gid_metype_bundle = list(metype_gids.values())
        gidvec = dry_run_distribution(gid_metype_bundle, stride, stride_offset)

        log_verbose("Loading node attributes... (subset of cells from each metype)")
        for gids in metype_gids.values():
            if not len(gids):
                continue
            gids = gids[:CELL_NODE_INFO_LIMIT]
            node_sel = libsonata.Selection(gids - 1)  # Load 0-based node ids
            morpho_names = node_pop.get_attribute("morphology", node_sel)
            mtypes = node_pop.get_attribute("mtype", node_sel)
            etypes = node_pop.get_attribute("etype", node_sel)
            model_templates = node_pop.get_attribute("model_template", node_sel)
            emodel_templates = [emodel.removeprefix("hoc:") for emodel in model_templates]
            meinfos.load_infoNP(gids, morpho_names, emodel_templates, mtypes, etypes)

        return gidvec, meinfos, total_cells

    def load_nodes_base_info():
        if SimConfig.dry_run or load_mode == "load_nodes_metype":
            return load_base_info_dry_run()

        meinfos = METypeManager()
        gidvec = split_round_robin(all_gids, stride, stride_offset, total_cells)

        if not len(gidvec):
            # Not enough cells to give this rank a few
            return gidvec, meinfos, total_cells

        log_verbose("Loading nodes info")
        node_sel = libsonata.Selection(gidvec - 1)  # 0-based node indices
        morpho_names = node_pop.get_attribute("morphology", node_sel)
        mtypes = node_pop.get_attribute("mtype", node_sel)
        try:
            etypes = node_pop.get_attribute("etype", node_sel)
        except libsonata.SonataError:
            logging.warning("etype not found in node population, setting to None")
            etypes = None
        model_templates = node_pop.get_attribute("model_template", node_sel)
        emodel_templates = [emodel.removeprefix("hoc:") for emodel in model_templates]
        if {"exc_mini_frequency", "inh_mini_frequency"}.issubset(attr_names):
            exc_mini_freqs = node_pop.get_attribute("exc_mini_frequency", node_sel)
            inh_mini_freqs = node_pop.get_attribute("inh_mini_frequency", node_sel)
        else:
            exc_mini_freqs = None
            inh_mini_freqs = None
        if {"threshold_current", "holding_current"}.issubset(dynamics_attr_names):
            threshold_currents = node_pop.get_dynamics_attribute("threshold_current", node_sel)
            holding_currents = node_pop.get_dynamics_attribute("holding_current", node_sel)
        else:
            threshold_currents = None
            holding_currents = None
        positions = np.array(
            [
                node_pop.get_attribute("x", node_sel),
                node_pop.get_attribute("y", node_sel),
                node_pop.get_attribute("z", node_sel),
            ]
        ).T
        rotations = _get_rotations(node_pop, node_sel)

        # For Sonata and new emodel hoc template, we need additional attributes for building metype
        # TODO: validate it's really the emodel_templates var we should pass here, or etype
        add_params_list = (
            None
            if not has_extra_data
            else _getNeededAttributes(
                node_pop, circuit_conf.METypePath, emodel_templates, gidvec - 1
            )
        )

        meinfos.load_infoNP(
            gidvec,
            morpho_names,
            emodel_templates,
            mtypes,
            etypes,
            threshold_currents,
            holding_currents,
            exc_mini_freqs,
            inh_mini_freqs,
            positions,
            rotations,
            add_params_list,
        )
        return gidvec, meinfos, total_cells

    # If dynamic properties are not specified simply return early
    if not load_dynamic_props:
        return load_nodes_base_info()

    # Check properties exist, eventually removing prefix
    def validate_property(prop_name):
        if prop_name.startswith("@dynamics:"):
            actual_prop_name = prop_name[len("@dynamics:") :]  # remove prefix
            if actual_prop_name not in dynamics_attr_names:
                raise Exception(f"Required Dynamics property {prop_name} not present")
        elif prop_name not in attr_names:
            raise Exception(f"Required extra property {prop_name} not present")

    [validate_property(p) for p in load_dynamic_props]

    # All good. Lets start reading!
    gidvec, meinfos, fullsize = load_nodes_base_info()

    if SimConfig.dry_run:
        load_nodes = np.fromiter(meinfos.keys(), dtype="uint32") - 1
        node_sel = libsonata.Selection(load_nodes)
    else:
        node_sel = libsonata.Selection(gidvec - 1)  # 0-based node indices

    for prop_name in load_dynamic_props:
        log_verbose("Loading extra property: %s ", prop_name)
        if prop_name.startswith("@dynamics:"):
            prop_name = prop_name[len("@dynamics:") :]
            prop_data = node_pop.get_dynamics_attribute(prop_name, node_sel)
        else:
            prop_data = node_pop.get_attribute(prop_name, node_sel)
        for gid, val in zip(meinfos.keys(), prop_data):
            meinfos[gid].extra_attrs[prop_name] = val

    return gidvec, meinfos, fullsize


def _getNeededAttributes(node_reader, etype_path, emodels, gidvec):
    """Read additional attributes required by emodel templates global var <emodel>__NeededAttributes
    Args:
        node_reader: libsonata node population
        etype_path: Location of emodel hoc templates
        emodels: Array of emodel names
        gidvec: Array of 0-based cell gids
    """
    add_params_list = []
    for gid, emodel in zip(gidvec, emodels):
        Nd.h.load_file(ospath.join(etype_path, emodel) + ".hoc")  # hoc doesn't throw
        attr_names = getattr(Nd, emodel + "_NeededAttributes", None)  # format "attr1;attr2;attr3"
        vals = []
        if attr_names is not None:
            vals = [node_reader.get_dynamics_attribute(name, gid) for name in attr_names.split(";")]
        add_params_list.append(vals)
    return add_params_list


def _get_rotations(node_reader, selection):
    """Get quaternions to rotate the cells

    Args:
        node_reader: libsonata node population
        selection: libsonata selection

    Returns:
        double vector of size [N][4] with the rotation quaternions in the order (x,y,z,w)
    """
    attr_names = node_reader.attribute_names
    if {"orientation_x", "orientation_y", "orientation_z", "orientation_w"}.issubset(attr_names):
        # Preferred way to present the rotation as quaternions
        return np.array(
            [
                node_reader.get_attribute("orientation_x", selection),
                node_reader.get_attribute("orientation_y", selection),
                node_reader.get_attribute("orientation_z", selection),
                node_reader.get_attribute("orientation_w", selection),
            ]
        ).T

    if {"rotation_angle_xaxis", "rotation_angle_yaxis", "rotation_angle_zaxis"}.intersection(
        attr_names
    ):
        # Some sonata nodes files use the Euler angle rotations, convert them to quaternions
        from scipy.spatial.transform import Rotation

        angle_x = (
            node_reader.get_attribute("rotation_angle_xaxis", selection)
            if "rotation_angle_xaxis" in attr_names
            else 0
        )
        angle_y = (
            node_reader.get_attribute("rotation_angle_yaxis", selection)
            if "rotation_angle_yaxis" in attr_names
            else 0
        )
        angle_z = (
            node_reader.get_attribute("rotation_angle_zaxis", selection)
            if "rotation_angle_yaxis" in attr_names
            else 0
        )
        euler_rots = np.array([angle_x, angle_y, angle_z]).T
        return Rotation.from_euler("xyz", euler_rots).as_quat()

    return None


@run_only_rank0
def _retrieve_unique_metypes(node_reader, all_gids, skip_metypes=()) -> dict:
    """Find unique mtype+emodel combinations in target to estimate resources in dry run.
    This function returns a list of lists of unique mtype+emodel combinations.
    Each of the inner lists contains gid for the same mtype+emodel combinations.

    Args:
        node_reader: node reader, libsonata only
        all_gids: list of all gids in target
    Returns:
        list of lists of unique mtype+emodel combinations
    """
    gidvec = np.array(all_gids)
    indexes = gidvec - 1
    if len(indexes) < 10:  # Ensure array is not too small (pybind11 #1392)
        indexes = indexes.tolist()

    if isinstance(node_reader, libsonata.NodePopulation):
        etypes = node_reader.get_attribute("etype", libsonata.Selection(indexes))
        mtypes = node_reader.get_attribute("mtype", libsonata.Selection(indexes))
    else:
        msg = f"Reader type {type(node_reader)} incompatible with dry run."
        raise TypeError(msg)

    gids_per_metype = defaultdict(list)
    count_per_metype = defaultdict(int)
    for gid, mtype, etype in zip(gidvec, mtypes, etypes):
        metype = f"{mtype}-{etype}"
        gids_per_metype[metype].append(gid)
        count_per_metype[metype] += 1

    logging.info(
        "Out of %d cells, found %d unique mtype+emodel combination",
        len(gidvec),
        len(gids_per_metype),
    )
    for metype, gid_list in gids_per_metype.items():
        logging.debug(
            "METype: %-20s instances: %-8d gids: %s", metype, len(gid_list), gid_list[:10]
        )

    # If the list is longer than 50, truncate it to 50 elements.
    # If the metype is already computed, skip it
    gid_metype_instantiate = {}
    for metype, gid_list in gids_per_metype.items():
        if metype not in skip_metypes:
            gid_metype_instantiate[metype] = np.array(gid_list, dtype="uint32")
        else:
            log_verbose("Skipping METype '%s' since it's already known", metype)

    return gid_metype_instantiate, count_per_metype



"""Implementation of Gid Sets with the ability of self offsetting and avoid
global overlapping
"""

from __future__ import annotations

from contextlib import contextmanager

import libsonata
import numpy as np

from . import MPI
from neurodamus.utils.pyutils import WeakList


class PopulationNodes:
    """Handle SelectionNodeSets belonging to a population. Given that Neuron doesnt
    inherently handle populations, we will have to apply gid offsetting.
    The class stores `SelectionNodeSet`s, and makes the required offsetting on-the-fly.

    This class is intended to be internal, since SelectionNodeSet instances can be
    freely created but only "apply" for offsetting when registered globally,
    in which case it delegates the processing to PopulationNodes.

    We store internal class-level _global_populations so that offsets are truly
    global, independently of the CellManager
    """

    _global_populations = []
    """Populations which may have offset"""
    _do_offsetting = True
    """False will freeze offsets to ensure final gids are consistent"""

    def __init__(self, name):
        """Ctor for a group of nodes belonging to the same population.
        It wont probably be used publicly given `get()` is also a factory.
        """
        self.name = name
        self.nodesets = WeakList()  # each population might contain several SelectionNodeSet's
        self.max_gid = 0  # maximum raw gid (without offset)
        self.offset = 0

    def _append(self, nodeset):
        self.nodesets.append(nodeset)
        self._update(nodeset)
        return self

    def _update(self, updated_nodeset):
        updated_nodeset._offset = self.offset
        if not self._do_offsetting:
            return
        local_max = max(self.max_gid, updated_nodeset._max_gid)
        max_gid = int(MPI.allreduce(local_max, MPI.MAX))
        if max_gid > self.max_gid:
            self.max_gid = max_gid
            self._update_offsets()

    @classmethod
    def register(cls, population, nodeset, **create_kw):
        return cls.get(population, create=True, **create_kw)._append(nodeset)

    @classmethod
    def freeze_offsets(cls):
        cls._do_offsetting = False

    @classmethod
    def reset(cls):
        cls._global_populations.clear()
        cls._do_offsetting = True

    @classmethod
    def all(cls):
        return cls._global_populations

    @classmethod
    def get(cls, population_name, *, create=False, **create_kw):
        obj = next(filter(lambda x: x.name == population_name, cls._global_populations), None)
        if not obj and create:
            obj = cls.create_pop(population_name, **create_kw)
        return obj

    @classmethod
    def create_pop(cls, population_name):
        new_population = cls(population_name)
        cls._global_populations.append(new_population)
        cls._global_populations = sorted(cls._global_populations, key=lambda x: x.name)
        new_population._compute_offset(cls._find_previous(new_population))
        return new_population

    @classmethod
    def _find_previous(cls, cur_pop):
        prev_nodeset = None
        for obj in cls.all():
            if obj is cur_pop:
                return prev_nodeset
            prev_nodeset = obj
        return None

    def _compute_offset(self, prev_gidpop):
        offset = 0
        # This offset is gonna be the offset+max_gid of the previous population, round up
        if prev_gidpop is not None:
            cur_max = prev_gidpop.offset + prev_gidpop.max_gid
            # round up 1000's. GIDs are 1 based: Blocks [1-1000], [1001-2000]
            offset = ((cur_max - 1) // 1000 + 1) * 1000
        self.offset = offset
        # Update individual nodesets
        for nodeset in self.nodesets:  # nodeset is a weakref
            nodeset()._offset = offset

    def _update_offsets(self):
        """Update all global offsets after adding gids"""
        update = False
        prev_gidpop = None
        for gidpop in self.all():
            if gidpop is self:
                update = True
                prev_gidpop = gidpop
                continue
            if update:
                # We are in a subsequent nodeset - re-set offsetting
                gidpop._compute_offset(prev_gidpop)
            prev_gidpop = gidpop

    @classmethod
    @contextmanager
    def offset_freezer(cls):
        cls._do_offsetting = False
        yield
        cls._do_offsetting = True


class SelectionNodeSet:
    """Set of nodes with optional global registration and offset handling.

    A shim over libsonata.Selection with optional populations, offsets and MEtype metadata per gid

    Note: this class is 0/1 based agnostic except for from_zero_based_libsonata_selection
    """

    def __init__(self, gids=None, gid_info=None):
        """Init.

        Args:
            gids: The gids to handle
            gid_info: a map containing METype information about each cell.
                In v5 and v6 values are METypeItem's
            offset: offset of the gids. Used to set a SelectionNodeSet from another one
        """
        self._offset = 0
        self._max_gid = 0  # maximum raw gid (without offset)
        self._population_group = None  # register in a population so gids can be unique
        self._selection = libsonata.Selection([])  # raw
        self._gid_info = {}
        if isinstance(gids, libsonata.Selection):
            self.add_selection(gids, gid_info)
        else:
            self.add_gids(gids, gid_info)

    offset = property(lambda self: self._offset)
    max_gid = property(lambda self: self._max_gid)

    def __repr__(self):
        gids = self.gids(raw_gids=True)
        n = len(gids)
        return (
            f"SelectionNodeSet(n={n}, "
            f"offset={self.offset}, "
            f"population={self.population_name}, "
            f"raw gids={gids})"
        )

    def __len__(self):
        return self._selection.flat_size

    def __iter__(self):
        for start, stop in self._selection.ranges:
            yield from range(start, stop)

    def iter_cell_info(self, raw_gids=True):
        """Iterate over GIDs with optional offset and metadata"""
        offset_add = 0 if raw_gids else self._offset

        for gid in self:
            yield gid + offset_add, self._gid_info.get(gid)

    def selection(self, raw_gids):
        """Return the internal Selection, optionally, with an offset"""
        if raw_gids:
            return self._selection

        return libsonata.Selection(
            [(start + self._offset, stop + self._offset) for start, stop in self._selection.ranges]
        )

    def gids(self, raw_gids):
        """Return all GIDs as a flat array, optionally offset by the population"""
        return np.asarray(self.selection(raw_gids=raw_gids).flatten(), dtype="uint32")

    def register_global(self, population_name):
        """Register this nodeset in a global population group

        Args:
            population_name: The name of the population these ids belong to
        """
        self._population_group = PopulationNodes.register(population_name, self)
        return self

    @property
    def population_name(self):
        return self._population_group.name if self._population_group else None

    def _check_update_offsets(self):
        """Check/reset offsets based on the other populations"""
        if self._population_group:
            self._population_group._update(self)  # Note: triggers a reduce.

    @classmethod
    def from_zero_based_libsonata_selection(cls, sel):
        """Create a nodeset from a 0-based libsonata.Selection to a 1-based SelectionNodeSet"""
        if not isinstance(sel, libsonata.Selection):
            raise TypeError(f"Expected libsonata.Selection, got {type(sel).__name__}")

        return cls(libsonata.Selection([(start + 1, stop + 1) for start, stop in sel.ranges]))

    def add_selection(self, selection: libsonata.Selection, gid_info=None):
        """Add libsonata.Selection GIDs and optional metadata, updating offsets and max_gid

        Args:
            selection: libsonata.Selection of GIDs
            gid_info: Optional map of GID to METype info (v5/v6 values are METypeItem)
        """
        if selection is None:
            return
        self._selection |= selection
        if self:
            # libsonata.Selection.ranges may be unsorted
            # Probably not needed since add_gids sorts
            self._max_gid = max(self.max_gid, np.max([i - 1 for _, i in self._selection.ranges]))
        if gid_info:
            self._gid_info.update(gid_info)
        self._check_update_offsets()  # check offsets (uses reduce)

    def add_gids(self, gids: list[int], gid_info=None):
        """Add GIDs and optional metadata, updating offsets and max_gid

        Args:
            gids: GIDs to add (list)
            gid_info: Optional map of GID to METype info (v5/v6 values are METypeItem)
        """
        if gids is None:
            return
        self.add_selection(selection=libsonata.Selection(gids), gid_info=gid_info)

    def intersection(self, other: SelectionNodeSet, raw_gids=False) -> libsonata.Selection:
        """Return libsonata.Selection in common with another nodeset

        For nodesets to intersect they must belong to the same population and
        have common gids
        """
        if not isinstance(other, SelectionNodeSet):
            raise TypeError(f"Expected SelectionNodeSet, got {type(other).__name__}")
        if self.population_name != other.population_name:
            return libsonata.Selection([])

        ans = self._selection & other._selection
        if raw_gids:
            return ans
        return libsonata.Selection(
            [(start + self.offset, stop + self.offset) for start, stop in ans.ranges]
        )

    def intersects(self, other):
        """Check if the current nodeset intersects another

        For nodesets to intersect they must belong to the same population and
        have common gids
        """
        return bool(self.intersection(other))

    def clear_cell_info(self):
        """Clear all stored GID metadata"""
        self._gid_info = None



"""Function for dumping cell state from NEURON context"""

import json
import logging

FILTER_SYNAPSE_ATTRS = ["delay_times", "delay_weights", "rng"]


def dump_cellstate(pc, cvode, gid, outputfile="cellstate.json"):
    """Dump cell, synapse, netcon states from NEURON context
    Args:
        pc: NEURON parallel contect
        cvode: NEURON CVode context, to get netcons list
        gid: cell gid in NEURON context
    """
    logging.info("Dump cell state for id %s", gid)
    cell = pc.gid2cell(gid)
    name = cell.hname()
    # remove the cell index from names
    cell_name = name[: name.find("[")]
    cell_prefix = name + "."
    res = {cell_name: {"gid": gid}}
    res[cell_name].update(dump_cell(cell))
    res[cell_name]["synapses"] = dump_synapses(cell.synlist, remove_prefix=cell_prefix)
    nclist = cvode.netconlist("", cell, "")
    res[cell_name]["n_netcons"] = nclist.count()
    res[cell_name]["netcons"] = dump_netcons(nclist, remove_prefix=cell_prefix)

    with open(outputfile, "w", encoding="utf-8") as f:
        json.dump(res, f, indent=2)


def dump_cell(cell) -> dict:
    """Dump cell state from NEURON context
    Args:
        cell: NEURON cell object
    """
    res = _read_object_attrs(cell)
    res["sections"] = []
    cell_name = cell.hname()
    for sec in cell.all:
        res_sec = {}
        sec_name = sec.hname()
        sec_name = sec_name.replace(cell_name + ".", "")
        res_sec["name"] = sec_name
        res_sec.update(_read_object_attrs(sec))
        res_sec["segments"] = []
        for seg in sec:
            attrs = {"x": seg.x}
            attrs.update(_read_object_attrs(seg))
            for key, item in attrs.items():
                # item is likely an nrn.Mechanism object
                if not isinstance(item, (int, float, str, list, dict, set)):
                    vals = _read_object_attrs(item)
                    attrs[key] = vals
            res_sec["segments"].append(attrs)
        res["sections"].append(res_sec)

    res["n_synapses"] = cell.synlist.count()
    return res


def dump_synapses(synlist, remove_prefix) -> list:
    """Dump states of synapses
    Args:
        synlist: Hoc list of synapse objects
        remove_prefix: remove prefix from key name
    """
    res = []
    for syn in synlist:
        attrs = {"name": syn.hname()}
        attrs.update(_read_object_attrs(syn, FILTER_SYNAPSE_ATTRS))
        attrs["location"] = syn.get_loc()
        attrs["segment"] = str(syn.get_segment()).removeprefix(remove_prefix)
        res.append(attrs)
    return res


def dump_netcons(nclist, remove_prefix) -> list:
    """Dump states of netcons from NEURON CVode
    Args:
        nclist: NEURON netcons list
        remove_prefix: remove prefix from key name
    """
    res = []
    for nc in nclist:
        attrs = {"name": nc.hname()}
        tsyn = nc.syn()
        attrs["target_syn"] = str(tsyn).removeprefix(remove_prefix)
        attrs["target_syn_segment"] = str(tsyn.get_segment()).removeprefix(remove_prefix)
        attrs["afferent_section_position"] = tsyn.get_segment().x
        attrs["srcgid"] = nc.srcgid()
        attrs["active"] = nc.active()
        attrs["weight"] = nc.weight[0]
        attrs.update(_read_object_attrs(nc))
        res.append(attrs)
    return res


def _read_object_attrs(obj, filter_keys=None):
    res = {}
    for x in dir(obj):
        if (
            (not filter_keys or x not in filter_keys)
            and not x.startswith("__")
            and not callable(getattr(obj, x))
        ):
            attr = getattr(obj, x)
            if isinstance(attr, float):
                attr = round(attr, 14)
            res[x] = attr
    return res



"""Py Wrappers for the HOC RNGs"""

from . import Neuron

# NOTE: These are pure wrappers, in the sense they don't create Python objects. Instead
#       Neuron objects are returned (using __new__)


class RNG:
    def __new__(cls, **kw):
        """Creates a default RNG (Random123 with ids of 0,1,2)"""
        seed = kw.get("seed", 0)
        return Random123(0, 1, 2, seed)

    @classmethod
    def create(cls, ids, seed=None):
        # type: (tuple, object) -> RNG
        assert len(ids) == 3, "Random123 requires three ids (as a tuple)"
        obj = Random123(ids[0], ids[1], ids[2], seed)

        return obj


class Random123(RNG):
    def __new__(cls, id1, id2, id3, seed=None):
        rng = Neuron.h.Random() if seed is None else Neuron.h.Random(seed)
        if seed is not None:
            rng.Random123_globalindex(seed)
        rng.Random123(id1, id2, id3)
        return rng


# Gamma-distributed sample generator (not available in NEURON)
def gamma(rng, a, b, N=1):  # noqa: N803
    """Sample N variates from a gamma distribution with parameters shape = a, scale = b
    using the NEURON random number generator rng.
    Uses the algorithm by Marsaglia and Tsang 2001.
    """
    from math import log, sqrt

    if a < 1:
        rng.uniform(0, 1)
        w = Neuron.h.Vector(N)
        w.setrand(rng)
        w.pow(1 / a)
        return gamma(rng, 1 + a, b, N).mul(w)

    d = a - 1 / 3
    c = 1 / 3 / sqrt(d)

    vec = Neuron.h.Vector(N)
    for i in range(N):
        while True:
            x = rng.normal(0, 1)
            v = 1 + c * x
            if v > 0:
                v = v * v * v
                u = rng.uniform(0, 1)
                if u < 1 - 0.0331 * x * x * x * x:
                    vec.x[i] = b * d * v
                    break
                if log(u) < 0.5 * x * x + d * (1 - v + log(v)):
                    vec.x[i] = b * d * v
                    break

    return vec



def extract_arguments(args):
    """Get the options for neurodamus and launch it.

    We can't use positional arguments with special so we look for
    --configFile=FILE, which defaults to simulation_config.json
    """
    first_argument_pos = 1
    init_py_reached = False
    config_file = "simulation_config.json"
    for i, arg in enumerate(args):
        if not init_py_reached:
            if arg.endswith("init.py"):
                first_argument_pos = i + 1
                init_py_reached = True
            elif arg.startswith("--configFile="):
                raise ValueError("Usage: ... init.py --configFile <config_file.json> ...")
            continue

        if not arg.startswith("-"):
            raise ValueError(
                "Positional arguments are not supported by init.py. "
                f"Found positional argument: '{arg}'"
            )
        if arg.startswith("--configFile="):
            config_file = arg.split("=")[1]

    result_args = [config_file] + [
        x for x in args[first_argument_pos:] if not x.startswith("--configFile=")
    ]
    return result_args



"""Collection of generic Python utilities."""

from __future__ import annotations

import subprocess  # noqa: S404
import weakref
from bisect import bisect_left
from enum import EnumMeta, IntEnum

import numpy as np


class StrEnumBase(IntEnum):
    __mapping__: list[tuple[str, int]] = []
    # default for when there is value. Leaving None throws an error
    __default__ = None
    # default when the string is not found in the mapping
    __invalid__ = None

    @classmethod
    def from_string(cls, s: str):
        if not s:
            return cls(cls.__default__)
        mapping = dict(cls.__mapping__)
        return cls(mapping.get(s.lower(), cls.__invalid__ if cls.__invalid__ is not None else s))

    def to_string(self) -> str:
        reverse__mapping__ = {v: k for k, v in self.__mapping__}
        return reverse__mapping__[self]

    def __str__(self):
        return f"{self.__class__.__name__}.{self.name}"

    @classmethod
    def default(cls):
        return cls(cls.__default__)


class CumulativeError(Exception):
    def __init__(self, errors=None):
        self.errors = errors or []
        self.is_error_appended = False

    def append(self, func_name, err):
        self.errors.append((func_name, err))
        self.is_error_appended = True

    def __bool__(self):
        return bool(self.errors)

    def __str__(self):
        if not self.errors:
            return "No errors."
        messages = [f"{func_name}: {type(err).__name__} -> {err}" for func_name, err in self.errors]
        return "Operation failed with multiple errors:\n" + "\n".join(messages)

    def raise_if_any(self):
        if self:
            raise self


def dict_filter_map(dic, mapp):
    """Filters a dict and converts the keys according to a given map"""
    return {mapp[key]: val for key, val in dic.items() if key in mapp}


def docopt_sanitize(docopt_opts):
    """Sanitizes docopt parsed key names"""
    opts = {}
    for key, val in docopt_opts.items():
        key = key.strip("<>-").replace("-", "_")
        if isinstance(val, str):
            if val.lower() in {"off", "false"}:
                val = False
            elif val.lower() in {"on", "true"}:
                val = True
        opts[key] = val
    return opts


class WeakList(list):  # noqa: FURB189
    def append(self, item):
        list.append(self, weakref.ref(item, self.remove))


class ConfigT:
    """Base class for configurations.

    This class serves as a base for set of configurations.
    By inheriting and setting several class-level attributes, instances will
    be able to initialize from kwargs and dictionaries with the same keys,
    effectively working as validators of fields with default values.
    Furthermore, for validation of values, the attributes may be Enums.

    ::\n
     class RunConfig(ConfigT):
        # NOTE: Enum fields: the FIRST value is the default
        mode = Enum("Mode", "BUILD_SIMULATE BUILD_ONLY")
        model_path = None
    """

    class _ConfigFlag:
        """A lightweith internal class to create flags"""

        __slots__ = ()

    REQUIRED = _ConfigFlag()

    def __init__(self, opt_dict=None, **opts):
        opt_dict = opt_dict or {}
        opt_dict.update(opts)
        self._init(self, opt_dict)

    @classmethod
    def _init(cls, obj, opts):
        for name, value in opts.items():
            if value is not None and not name.startswith("_") and hasattr(obj, name):
                default = getattr(obj, name)
                if type(default) is EnumMeta:
                    value = default[value]  # enum as validator
                setattr(obj, name, value)

        for name, value in cls.__dict__.items():
            if name not in obj.__dict__ and (value is cls.REQUIRED or type(value) is EnumMeta):
                raise ValueError(f"Config field {name} is mandatory")

        obj._all = opts

    # dict interface for compat
    def __setitem__(self, name, value):
        self._all[name] = value

    def __getitem__(self, name):
        return self._all[name]

    def get(self, *args):
        return self._all.get(*args)

    def __contains__(self, name):
        return name in self._all

    all = property(lambda self: self._all)

    def as_dict(self, subset=None, excludes=()):
        return {
            key: val
            for key, val in vars(self).items()
            if val is not None
            and not key.startswith("_")
            and key not in excludes
            and (subset is None or key in subset)
        }


def bin_search(container, key, keyf=None):
    """Performs binary search in a container, retrieving the index where key should be inserted
    to keep ordering. Accepts a key function to be applied to each element of the container.

    Args:
        container: The container to be searched through
        key: The key to look for
        keyf: (Optional) the function transforming container elements into comparable keys

    Returns: The position where the element is to be inserted to keep ordering.

    """
    if keyf is None:
        return bisect_left(container, key)

    binsrch_low = 0
    binsrch_high = len(container)

    while binsrch_low < binsrch_high:
        binsrch_mid = int((binsrch_low + binsrch_high) * 0.5)
        if key > keyf(container[binsrch_mid]):
            binsrch_low = binsrch_mid + 1
        else:
            binsrch_high = binsrch_mid
    return binsrch_low


class ConsoleColors:
    """Helper class for formatting console text."""

    BLACK, RED, GREEN, YELLOW, BLUE, MAGENTA, CYAN, WHITE, _, DEFAULT = range(30, 40)
    NORMAL, BOLD, DIM, UNDERLINED, BLINK, INVERTED, HIDDEN = [a << 8 for a in range(7)]

    # These are the sequences needed to control output
    _CHANGE_SEQ = "\033[{}m"
    _RESET_SEQ = "\033[0m"

    @classmethod
    def reset(cls):
        return cls._RESET_SEQ

    @classmethod
    def format_text(cls, text, color, style=None):
        style = (style or color) >> 8
        format_seq = str(color & 0x00FF) + ((";" + str(style)) if style else "")
        return cls._CHANGE_SEQ.format(format_seq) + text + cls._RESET_SEQ


def append_recarray(target_array, record):
    """Append a np.record to a np.recarray"""
    if target_array is None:
        target_array = np.recarray(1, dtype=record.dtype)
        target_array[0] = record
    elif not isinstance(target_array, np.recarray) or target_array.dtype != record.dtype:
        raise TypeError("Can not append a recode with a different dtype to the target array")
    else:
        nrows = target_array.shape[0]
        target_array.resize(nrows + 1, refcheck=False)
        target_array[nrows] = record
    return target_array


def gen_ranges(limit, blocklen, low=0, block_increase_rate=1):
    """Generates ranges in block intervals for a given length
    block_increase_rate may be >1 in case we want the block to get increasingly large
    """
    while low < limit:
        high = min(low + blocklen, limit)
        yield low, high
        low = high
        blocklen = int(blocklen * block_increase_rate)


def rmtree(path):
    """Recursively remove a path

    Note: shutils.rmtree wouldn't complete for directories with many files.
    See:
    https://github.com/openbraininstitute/neurodamus/pull/247/files/e9d12100b22bf512fdcd624022d9d999cb50db77#r2079776328  # noqa: E501
    """  # noqa: E501
    subprocess.call(["/bin/rm", "-rf", path])  # noqa: S603


def cache_errors(func):
    """Decorator that catches exceptions and appends
    (func name, exception) to `cumulative_error` if provided.
    """

    def wrapper(*args, cumulative_error: CumulativeError | None = None, **kwargs):
        try:
            return func(*args, **kwargs)
        except Exception as e:
            if cumulative_error is not None:
                cumulative_error.append(func.__name__, e)
            else:
                raise
            return None

    return wrapper



from __future__ import annotations

import logging
import struct
from collections.abc import Iterable
from dataclasses import dataclass, field, fields
from pathlib import Path
from typing import Union, get_args, get_origin, get_type_hints

from ._utils import run_only_rank0
from neurodamus.report_parameters import ReportType


@dataclass
class CoreReportConfigEntry:
    report_name: str
    target_name: str
    report_type: str
    report_variable: str
    unit: str
    report_format: str
    sections: str
    compartments: str
    dt: float
    start_time: float
    end_time: float
    buffer_size: int
    scaling: str

    # public lists
    gids: list[int] = field(default_factory=list, init=False)
    points_section_ids: list[int] = field(default_factory=list, init=False)
    points_compartment_ids: list[int] = field(default_factory=list, init=False)

    def __post_init__(self):
        # type coercion and path resolution
        hints = get_type_hints(self)
        for f in fields(self):
            val = getattr(self, f.name)
            coerced = self._coerce_type(hints[f.name], val)
            setattr(self, f.name, coerced)

    @staticmethod
    def _coerce_type(typ, val):
        result = None

        typ = get_origin(typ) or typ
        if typ is Union:
            typ = next(t for t in get_args(typ) if t is not type(None))

        if val is not None:
            if isinstance(val, typ):
                result = val
            elif typ is bool:
                result = val.strip().lower() in {"true", "1"} if isinstance(val, str) else bool(val)
            elif typ is int:
                result = int(val)
            elif typ is float:
                result = float(val)
            elif typ is str:
                result = str(val)
            else:
                result = val

        return result

    @property
    def num_gids(self):
        return len(self.gids)

    @run_only_rank0
    def set_gids(self, gids):
        assert self.report_type != "compartment_set"
        assert isinstance(gids, Iterable)
        assert len(gids)
        self.gids = gids

    @run_only_rank0
    def set_points(self, gids, section_ids, compartment_ids):
        assert self.report_type == "compartment_set"
        assert isinstance(gids, Iterable)
        assert isinstance(section_ids, Iterable)
        assert isinstance(compartment_ids, Iterable)
        assert len(gids)
        assert len(gids) == len(section_ids) == len(compartment_ids), (
            f"All input lists must have the same length, "
            f"got gids: {len(gids)}, section_ids: {len(section_ids)}, "
            f"compartment_ids: {len(compartment_ids)}"
        )
        self.gids = gids
        self.points_section_ids = section_ids
        self.points_compartment_ids = compartment_ids

    @run_only_rank0
    def dump(self, f):
        assert len(self.gids)

        # text line with field values
        line_values = [str(getattr(self, f.name)) for f in fields(self) if f.init]
        line_values.insert(11, str(self.num_gids))
        f.write((" ".join(line_values) + "\n").encode())

        # binary gids
        f.write(struct.pack(f"{len(self.gids)}i", *self.gids))
        f.write(b"\n")

        # binary points
        if self.points_section_ids:
            f.write(struct.pack(f"{len(self.points_section_ids)}i", *self.points_section_ids))
            f.write(b"\n")
        if self.points_compartment_ids:
            f.write(
                struct.pack(f"{len(self.points_compartment_ids)}i", *self.points_compartment_ids)
            )
            f.write(b"\n")

    @staticmethod
    def _get_binary_int_array(f, num_elements):
        data = f.read(num_elements * 4)
        if len(data) != num_elements * 4:
            raise ValueError(f"Expected {num_elements * 4} bytes, got {len(data)}")
        f.readline()
        return list(struct.unpack(f"{num_elements}i", data))

    @classmethod
    def load_from_file(cls, f):
        line = f.readline()
        if not line:
            return None
        tokens = line.decode().strip().split()
        num_gids = int(tokens.pop(11))
        entry = cls(*tokens)
        gids = cls._get_binary_int_array(f, num_gids)
        if entry.report_type == "compartment_set":
            section_ids = cls._get_binary_int_array(f, num_gids)
            compartment_ids = cls._get_binary_int_array(f, num_gids)
            entry.set_points(gids, section_ids, compartment_ids)
        else:
            entry.set_gids(gids)
        return entry

    @classmethod
    def from_report_params(cls, rep_params):
        entry = cls(
            report_name=rep_params.name,
            target_name=rep_params.target.name,
            report_type=rep_params.type.to_string(),
            report_variable=",".join(rep_params.report_on.split()),
            unit=rep_params.unit,
            report_format=rep_params.format,
            sections=rep_params.sections.to_string(),
            compartments=rep_params.compartments.to_string(),
            dt=rep_params.dt,
            start_time=rep_params.start,
            end_time=rep_params.end,
            buffer_size=rep_params.buffer_size,
            scaling=rep_params.scaling.to_string(),
        )
        if rep_params.type == ReportType.COMPARTMENT_SET:
            gids = [i.gid for i in rep_params.points for _section_id, _sec, _x in i]
            section_ids = [section_id for i in rep_params.points for section_id, _sec, _x in i]
            compartment_ids = [
                sec.sec(x).node_index() for i in rep_params.points for _section_id, sec, x in i
            ]
            entry.set_points(gids, section_ids, compartment_ids)
        else:
            entry.set_gids(rep_params.target.gids(raw_gids=False))
        return entry


@dataclass
class CoreReportConfig:
    """Handler of report.conf, the configuration of the coreNeuron reports."""

    reports: dict[str, CoreReportConfigEntry] = field(default_factory=dict, init=False)
    pop_offsets: dict[str, int] = field(default=None, init=False)
    spike_filename: str = field(default=None, init=False)

    def add_entry(self, entry: CoreReportConfigEntry):
        logging.info(
            "Adding report %s for CoreNEURON with %s gids", entry.report_name, entry.num_gids
        )
        self.reports[entry.report_name] = entry

    def set_pop_offsets(self, pop_offsets: dict[str, int]):
        """Set the population offsets for the reports (skip None key)."""
        assert isinstance(pop_offsets, dict), "pop_offsets must be a dictionary"
        self.pop_offsets = {k: v for k, v in pop_offsets.items() if k is not None}

    def set_spike_filename(self, spike_path: str):
        """Set the spike filename for the reports."""
        if spike_path is not None:
            # Get only the spike file name
            self.spike_filename = spike_path.rsplit("/", maxsplit=1)[-1]

    @run_only_rank0
    def dump(self, path: str | Path):
        path = Path(path).resolve()
        path.parent.mkdir(parents=True, exist_ok=True)
        # always override
        logging.info("Writing coreneuron report config file: %s", path)
        with open(path, "wb") as f:
            # number of reports
            f.write(f"{len(self.reports)}\n".encode())
            # dump each entry
            for entry in self.reports.values():
                entry.dump(f)

            f.write(f"{len(self.pop_offsets)}\n".encode())
            for k, v in self.pop_offsets.items():
                if v is not None:
                    f.write(f"{k} {v}\n".encode())
                else:
                    f.write(f"{k}\n".encode())

            if self.spike_filename:
                f.write(f"{self.spike_filename}\n".encode())
        logging.info("Done! coreneuron report config was written")

    @classmethod
    def load(cls, path: str) -> CoreReportConfig:
        config = cls()
        with open(path, "rb") as f:
            # --- Read number of reports ---
            line = f.readline()
            if not line:
                return config  # empty file
            try:
                num_reports = int(line.decode().strip())
            except ValueError as err:
                raise ValueError(f"Invalid number of reports in file: {line}") from err

            # --- Read each report ---
            for _ in range(num_reports):
                entry = CoreReportConfigEntry.load_from_file(f)
                if entry is None:
                    raise ValueError("Unexpected EOF while reading reports")
                config.add_entry(entry)

            # --- Mandatory pop_offsets ---
            line = f.readline()
            if not line:
                raise ValueError("Missing population offsets")
            pop_count = int(line.decode().strip())
            pop_offsets = {}
            for _ in range(pop_count):
                line = f.readline()
                if not line:
                    break
                parts = line.decode().strip().split()
                k = parts[0]
                v = int(parts[1]) if len(parts) > 1 else None
                pop_offsets[k] = v
            config.set_pop_offsets(pop_offsets)

            # --- Optional spike filename ---
            line = f.readline()
            if line:
                spike_filename = line.decode().strip()
                if spike_filename:
                    config.set_spike_filename(spike_filename)

        return config

    @staticmethod
    @run_only_rank0
    def update_file(file_path: str, substitutions: dict[str, dict[str, int]]):
        """Update a report configuration file by applying attribute substitutions to
        one or more reports.

        Args:
            file_path (str): Path to the configuration file to load and update.
            substitutions (dict[str, dict[str, int]]): A mapping from report names
                to dictionaries of attribute-value pairs to be updated.

        Example:
            >>> substitutions = {"r1": {"buffer_size": 11}}
            >>> CoreReportConfig.update_file("config.yaml", substitutions)
            # Updates report 'r1' so that r1.buffer_size == 11
        """
        conf = CoreReportConfig.load(file_path)
        for report_name, targets in substitutions.items():
            report = conf.reports[report_name]
            for attr, new_val in targets.items():
                if not hasattr(report, attr):
                    raise AttributeError(f"Missing attribute '{attr}' in {report!r}")

                current_val = getattr(report, attr)
                if not isinstance(new_val, type(current_val)):
                    raise TypeError(
                        f"Type mismatch for '{attr}': expected {type(current_val).__name__}, "
                        f"got {type(new_val).__name__}. "
                        f"Current value={current_val!r}, attempted new value={new_val!r}"
                    )

                setattr(report, attr, new_val)
        conf.dump(file_path)



# https://bbpteam.epfl.ch/project/spaces/display/BGLIB/Neurodamus
# Copyright 2005-2021 Blue Brain Project, EPFL. All rights reserved.
"""Implements coupling artificial stimulus into simulation

New Stimulus classes must be registered, using the appropriate decorator.
Also, when instantiated by the framework, __init__ is passed three arguments
(1) target_points (2) stim_info: dict (3) cell_manager. Example

>>> @StimulusManager.register_type
>>> class ShotNoise:
>>>
>>> def __init__(self, target_points, stim_info: dict, cell_manager):
>>>     for point in target_points:
>>>         gid = point.gid
>>>         cell = cell_manager.get_cell(gid)

"""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from .core import NeuronWrapper as Nd, random
from .core.configuration import ConfigurationError, SimConfig
from .core.stimuli import ConductanceSource, CurrentSource
from .utils.logging import log_verbose

if TYPE_CHECKING:
    from .target_manager import TargetPointList, TargetSpec
    from .utils import compat


class StimulusManager:
    """A manager for synaptic artificial Stimulus.
    Old stimulus resort to hoc implementation
    """

    _stim_types = {}  # stimulus handled in Python

    def __init__(self, target_manager):
        self._target_manager = target_manager
        self._stim_seed = SimConfig.run_conf.get("StimulusSeed")
        self._stimulus = []
        self.reset_helpers()  # reset helpers for multi-cycle builds

    def interpret(self, target_spec, stim_info):
        stim_t = self._stim_types.get(stim_info["Pattern"])
        if not stim_t:
            msg = f"No implementation for Stimulus {stim_info['Pattern']}"
            raise ConfigurationError(msg)
        if self._stim_seed is None and getattr(stim_t, "IsNoise", False):
            logging.warning(
                "StimulusSeed unset (default %d), set explicitly to vary noisy stimuli across runs",
                SimConfig.rng_info.getStimulusSeed(),
            )
        cell_manager = self._target_manager._cell_manager
        log_verbose("Interpret stimulus")
        target_points = self.get_point_list(target_spec, stim_info, cell_manager)
        stim = stim_t(target_points, stim_info, cell_manager)
        self._stimulus.append(stim)

    def get_point_list(
        self, target_spec: TargetSpec, stim_info: dict, cell_manager
    ) -> compat.List[TargetPointList]:
        """Return points from target, using a specified compartment set if given.

        Args:
            target_spec: Object providing points.
            stim_info: May contain "CompartmentSet".
            cell_manager: Necessary to retrieve references to neuron sections and neurons in general

        Returns:
            compat.List of TargetPointList (one TargetPointList per gid.
            In general it is a jagged array)
        """
        target = self._target_manager.get_target(target_spec)
        if compartment_set_name := stim_info.get("CompartmentSet"):
            compartment_set = self._target_manager.get_compartment_set(compartment_set_name)
            return target.get_point_list_from_compartment_set(
                cell_manager=cell_manager, compartment_set=compartment_set
            )
        return target.get_point_list(cell_manager=cell_manager)

    @staticmethod
    def reset_helpers():
        ShotNoise.stim_count = 0
        Noise.stim_count = 0
        OrnsteinUhlenbeck.stim_count = 0

    @classmethod
    def register_type(cls, stim_class):
        """Registers a new class as a handler for a new stim type"""
        cls._stim_types[stim_class.__name__] = stim_class
        return stim_class

    def saveStatePreparation(self, ss_obj):
        for stim in self._stimulus:
            ss_obj.ignore(stim)


class BaseStim:
    """Barebones stimulus class"""

    IsNoise = False

    def __init__(self, _target_points: list[TargetPointList], stim_info: dict, _cell_manager):
        self.duration = float(stim_info["Duration"])  # duration [ms]
        self.delay = float(stim_info["Delay"])  # start time [ms]
        self.represents_physical_electrode = stim_info.get("RepresentsPhysicalElectrode", False)


@StimulusManager.register_type
class OrnsteinUhlenbeck(BaseStim):
    """Ornstein-Uhlenbeck process, injected as current or conductance"""

    IsNoise = True
    stim_count = 0  # global count for seeding

    def __init__(self, target_points: list[TargetPointList], stim_info: dict, cell_manager):
        super().__init__(target_points, stim_info, cell_manager)

        self.stimList = []  # sources go here

        if not self.parse_check_all_parameters(stim_info):
            return  # nothing to do, stim is a no-op

        # setup random seeds
        seed1 = OrnsteinUhlenbeck.stim_count + 2997  # stimulus block seed
        seed2 = SimConfig.rng_info.getStimulusSeed() + 291204  # stimulus type seed
        seed3 = (lambda x: x + 123) if self.seed is None else (lambda _x: self.seed)  # GID seed

        # apply stim to each point in target_points
        for target_point_list in target_points:
            gid = target_point_list.gid
            cell = cell_manager.get_cell(gid)

            self.compute_parameters(cell)

            for sec_id, sc in enumerate(target_point_list.sclst):
                # skip sections not in this split
                if not sc.exists():
                    continue

                rng = random.Random123(seed1, seed2, seed3(gid))  # setup RNG
                ou_args = (self.tau, self.sigma, self.mean, self.duration)
                ou_kwargs = {
                    "dt": self.dt,
                    "delay": self.delay,
                    "rng": rng,
                    "represents_physical_electrode": self.represents_physical_electrode,
                }
                # inject Ornstein-Uhlenbeck signal
                if stim_info["Mode"] == "Conductance":
                    cs = ConductanceSource.ornstein_uhlenbeck(
                        *ou_args, **ou_kwargs, base_amp=self.reversal
                    )
                else:
                    cs = CurrentSource.ornstein_uhlenbeck(*ou_args, **ou_kwargs)
                # attach source to section
                cs.attach_to(sc.sec, target_point_list.x[sec_id])
                self.stimList.append(cs)  # save source

        OrnsteinUhlenbeck.stim_count += 1  # increment global count

    def parse_check_all_parameters(self, stim_info: dict):
        self.dt = float(stim_info.get("Dt", 0.25))  # stimulus timestep [ms]
        if self.dt <= 0:
            raise Exception(f"{self.__class__.__name__} time-step must be positive")

        self.reversal = float(stim_info.get("Reversal", 0.0))  # reversal potential [mV]

        if stim_info["Mode"] not in {"Current", "Conductance"}:
            raise Exception(
                f"{self.__class__.__name__} must be used with mode Current or Conductance"
            )

        self.tau = float(stim_info["Tau"])  # relaxation time [ms]
        if self.tau < 0:
            raise Exception(f"{self.__class__.__name__} relaxation time must be non-negative")

        # parse and check stimulus-specific parameters
        if not self.parse_check_stim_parameters(stim_info):
            return False  # nothing to do, stim is a no-op

        self.seed = stim_info.get("Seed")  # random seed override
        if self.seed is not None:
            self.seed = int(self.seed)

        return True

    def parse_check_stim_parameters(self, stim_info):
        self.sigma = float(stim_info["Sigma"])  # signal stdev [uS]
        if self.sigma <= 0:
            raise Exception(f"{self.__class__.__name__} standard deviation must be positive")

        self.mean = float(stim_info["Mean"])  # signal mean [uS]
        if self.mean < 0 and abs(self.mean) > 2 * self.sigma:
            logging.warning("%s signal is mostly zero", self.__class__.__name__)

        return True

    def compute_parameters(self, cell):
        # nothing to do
        pass


@StimulusManager.register_type
class RelativeOrnsteinUhlenbeck(OrnsteinUhlenbeck):
    """Ornstein-Uhlenbeck process, injected as current or conductance,
    relative to cell threshold current or inverse input resistance
    """

    IsNoise = True

    def __init__(self, target_points: list[TargetPointList], stim_info: dict, cell_manager):
        super().__init__(target_points, stim_info, cell_manager)

    def parse_check_stim_parameters(self, stim_info):
        self.mean_perc = float(stim_info["MeanPercent"])
        self.sigma_perc = float(stim_info["SDPercent"])

        if stim_info["Mode"] == "Current":
            self.get_relative = lambda x: x.getThreshold()
        else:
            self.get_relative = lambda x: 1.0 / x.input_resistance

        return True

    def compute_parameters(self, cell):
        # threshold current [nA] or inverse input resistance [uS]
        rel_prop = self.get_relative(cell)

        self.sigma = (self.sigma_perc / 100) * rel_prop  # signal stdev [nA or uS]
        if self.sigma <= 0:
            raise Exception(f"{self.__class__.__name__} standard deviation must be positive")

        self.mean = (self.mean_perc / 100) * rel_prop  # signal mean [nA or uS]
        if self.mean < 0 and abs(self.mean) > 2 * self.sigma:
            logging.warning("%s signal is mostly zero", self.__class__.__name__)

        return True


@StimulusManager.register_type
class ShotNoise(BaseStim):
    """ShotNoise stimulus handler implementing Poisson shot noise
    with bi-exponential response and gamma-distributed amplitudes
    """

    IsNoise = True
    stim_count = 0  # global count for seeding

    def __init__(self, target_points: list[TargetPointList], stim_info: dict, cell_manager):
        super().__init__(target_points, stim_info, cell_manager)

        self.stimList = []  # CurrentSource's go here

        if not self.parse_check_all_parameters(stim_info):
            return  # nothing to do, stim is a no-op

        # setup random seeds
        seed1 = ShotNoise.stim_count + 2997  # stimulus block seed
        seed2 = SimConfig.rng_info.getStimulusSeed() + 19216  # stimulus type seed
        seed3 = (lambda x: x + 123) if self.seed is None else (lambda _x: self.seed)  # GID seed

        # apply stim to each point in target_points
        for target_point_list in target_points:
            gid = target_point_list.gid
            cell = cell_manager.get_cell(gid)

            self.compute_parameters(cell)

            for sec_id, sc in enumerate(target_point_list.sclst):
                # skip sections not in this split
                if not sc.exists():
                    continue

                rng = random.Random123(seed1, seed2, seed3(gid))  # setup RNG
                shotnoise_args = (
                    self.tau_D,
                    self.tau_R,
                    self.rate,
                    self.amp_mean,
                    self.amp_var,
                    self.duration,
                )
                shotnoise_kwargs = {
                    "dt": self.dt,
                    "delay": self.delay,
                    "rng": rng,
                    "represents_physical_electrode": self.represents_physical_electrode,
                }
                # generate shot noise current source
                if stim_info["Mode"] == "Conductance":
                    cs = ConductanceSource.shot_noise(
                        *shotnoise_args, **shotnoise_kwargs, base_amp=self.reversal
                    )
                else:
                    cs = CurrentSource.shot_noise(*shotnoise_args, **shotnoise_kwargs)
                # attach current source to section
                cs.attach_to(sc.sec, target_point_list.x[sec_id])
                self.stimList.append(cs)  # save CurrentSource

        ShotNoise.stim_count += 1  # increment global count

    def parse_check_all_parameters(self, stim_info: dict):
        if stim_info["Mode"] not in {"Current", "Conductance"}:
            raise Exception(
                f"{self.__class__.__name__} must be used with mode Current or Conductance"
            )

        self.reversal = float(stim_info.get("Reversal", 0.0))  # reversal potential [mV]

        # time parameters
        self.dt = float(stim_info.get("Dt", 0.25))  # stimulus timestep [ms]
        if self.dt <= 0:
            raise Exception(f"{self.__class__.__name__} time-step must be positive")

        ntstep = int(self.duration / self.dt)  # number of timesteps [1]
        if ntstep == 0:
            return False  # nothing to do, stim is a no-op

        # bi-exponential parameters
        self.tau_R = float(stim_info["RiseTime"])  # rise time [ms]
        self.tau_D = float(stim_info["DecayTime"])  # decay time [ms]
        if self.tau_R >= self.tau_D:
            klass = self.__class__.__name__
            raise Exception(f"{klass} bi-exponential rise time must be smaller than decay time")

        # parse and check stimulus-specific parameters
        if not self.parse_check_stim_parameters(stim_info):
            return False  # nothing to do, stim is a no-op

        # seed
        self.seed = stim_info.get("Seed")  # random seed override
        if self.seed is not None:
            self.seed = int(self.seed)

        return True

    def parse_check_stim_parameters(self, stim_info: dict):
        """Parse parameters for ShotNoise stimulus"""
        # event rate of Poisson process [Hz]
        self.rate = float(stim_info["Rate"])

        # mean amplitude of shots [nA or uS]
        # when negative we invert the sign of the current
        self.amp_mean = float(stim_info["AmpMean"])
        if self.amp_mean == 0:
            raise Exception(f"{self.__class__.__name__} amplitude mean must be non-zero")

        # variance of amplitude of shots [nA^2 or uS^2]
        self.amp_var = float(stim_info["AmpVar"])
        if self.amp_var <= 0:
            raise Exception(f"{self.__class__.__name__} amplitude variance must be positive")

        return self.rate > 0  # no-op if rate == 0

    def compute_parameters(self, cell):
        # nothing to do
        pass

    def params_from_mean_sd(self, mean, sd):
        """Compute bi-exponential shot noise parameters from desired mean and std. dev. of signal.

        Analytical result derived from a generalization of Campbell's theorem present in
        Rice, S.O., "Mathematical Analysis of Random Noise", BSTJ 23, 3 Jul 1944.
        """
        from math import exp, log

        # bi-exponential time to peak [ms]
        t_peak = log(self.tau_D / self.tau_R) / (1 / self.tau_R - 1 / self.tau_D)
        # bi-exponential peak height [1]
        F_peak = exp(-t_peak / self.tau_D) - exp(-t_peak / self.tau_R)

        # utility constants
        Xi = (self.tau_D - self.tau_R) / F_peak
        A = 1 / (self.tau_D + self.tau_R)
        B = 1 / ((self.tau_D + 2 * self.tau_R) * (2 * self.tau_D + self.tau_R))

        # skewness
        skew_bnd_min = (8 / 3) * (B / A**2) * (sd / mean)
        skew = (1 + self.rel_skew) * skew_bnd_min
        if skew < skew_bnd_min or skew > 2 * skew_bnd_min:
            raise Exception(f"{self.__class__.__name__} skewness out of bounds")

        # cumulants
        lambda2_1 = sd**2 / mean  # lambda2 over lambda1
        lambda3_2 = sd * skew  # lambda3 over lambda2
        theta1pk = 2 / (A * Xi) * lambda2_1  # = (1 + k) * theta
        theta2pk = (3 * A) / (4 * B * Xi) * lambda3_2  # = (2 + k) * theta

        # derived parameters
        self.amp_mean = 2 * theta1pk - theta2pk  # mean amplitude [nA or uS]
        self.amp_var = self.amp_mean * (theta2pk - theta1pk)  # variance of amplitude [nA^2 or uS^2]
        rate_ms = mean / (self.amp_mean * Xi)  # event rate in 1 / ms
        self.rate = rate_ms * 1000  # event rate in 1 / s [Hz]


@StimulusManager.register_type
class RelativeShotNoise(ShotNoise):
    """RelativeShotNoise stimulus handler, same as ShotNoise
    but parameters relative to cell threshold current or inverse input resistance
    """

    IsNoise = True

    def __init__(self, target_points: list[TargetPointList], stim_info: dict, cell_manager):
        super().__init__(target_points, stim_info, cell_manager)

    def parse_check_stim_parameters(self, stim_info: dict):
        """Parse parameters for RelativeShotNoise stimulus"""
        # signal mean as percent of cell's threshold [1],
        # when negative we invert the sign of the current
        self.mean_perc = float(stim_info["MeanPercent"])

        # signal standard deviation as percent of cell's threshold [1]
        self.sd_perc = float(stim_info["SDPercent"])
        if self.sd_perc <= 0:
            raise Exception(f"{self.__class__.__name__} stdev percent must be positive")
        if self.sd_perc < 1:
            logging.warning(
                "%s stdev percent too small gives a very high event rate", self.__class__.__name__
            )

        # relative skewness of signal as a [0,1] fraction [1]
        self.rel_skew = float(stim_info.get("RelativeSkew", 0.5))
        if self.rel_skew < 0.0 or self.rel_skew > 1.0:
            raise Exception(f"{self.__class__.__name__} relative skewness must be in [0,1]")

        if stim_info["Mode"] == "Current":
            self.get_relative = lambda x: x.getThreshold()
        else:
            self.get_relative = lambda x: 1.0 / x.input_resistance

        return self.mean_perc != 0  # no-op if mean_perc == 0

    def compute_parameters(self, cell):
        # threshold current [nA] or inverse input resistance [uS]
        rel_prop = self.get_relative(cell)
        mean = (self.mean_perc / 100) * rel_prop  # desired mean [nA or uS]
        sd = (self.sd_perc / 100) * rel_prop  # desired standard deviation [nA or uS]
        super().params_from_mean_sd(mean, sd)


@StimulusManager.register_type
class AbsoluteShotNoise(ShotNoise):
    """AbsoluteShotNoise stimulus handler, same as ShotNoise
    but parameters from given mean and std. dev.
    """

    IsNoise = True

    def __init__(self, target_points: list[TargetPointList], stim_info: dict, cell_manager):
        super().__init__(target_points, stim_info, cell_manager)

    def parse_check_stim_parameters(self, stim_info: dict):
        """Parse parameters for AbsoluteShotNoise stimulus"""
        # signal mean [nA]
        self.mean = float(stim_info["Mean"])

        # signal standard deviation [nA]
        self.sd = float(stim_info["Sigma"])
        if self.sd <= 0:
            raise Exception(f"{self.__class__.__name__} stdev must be positive")

        # relative skewness of signal as a [0,1] fraction [1]
        self.rel_skew = float(stim_info.get("RelativeSkew", 0.5))
        if self.rel_skew < 0.0 or self.rel_skew > 1.0:
            raise Exception(f"{self.__class__.__name__} relative skewness must be in [0,1]")

        return True

    def compute_parameters(self, cell):
        super().params_from_mean_sd(self.mean, self.sd)


@StimulusManager.register_type
class Linear(BaseStim):
    """Injects a linear current ramp."""

    def __init__(self, target_points: list[TargetPointList], stim_info: dict, cell_manager):
        super().__init__(target_points, stim_info, cell_manager)

        self.stimList = []  # CurrentSource's go here

        if not self.parse_check_all_parameters(stim_info):
            return  # nothing to do, stim is a no-op

        # apply stim to each point in target_points
        for target_point_list in target_points:
            gid = target_point_list.gid
            cell = cell_manager.get_cell(gid)

            self.compute_parameters(cell)

            for sec_id, sc in enumerate(target_point_list.sclst):
                # skip sections not in this split
                if not sc.exists():
                    continue

                # generate ramp current source
                cs = CurrentSource.ramp(
                    self.amp_start,
                    self.amp_end,
                    self.duration,
                    delay=self.delay,
                    represents_physical_electrode=self.represents_physical_electrode,
                )
                # attach current source to section
                cs.attach_to(sc.sec, target_point_list.x[sec_id])
                self.stimList.append(cs)  # save CurrentSource

    def parse_check_all_parameters(self, stim_info: dict):
        # Amplitude at start
        self.amp_start = float(stim_info["AmpStart"])

        # Amplitude at end (optional, else same as start)
        self.amp_end = float(stim_info.get("AmpEnd", self.amp_start))

        return self.amp_start != 0 or self.amp_end != 0  # no-op if both 0

    def compute_parameters(self, cell):
        pass  # nothing to do


@StimulusManager.register_type
class Hyperpolarizing(Linear):
    """Injects a constant step with a cell's hyperpolarizing current."""

    def __init__(self, target_points: list[TargetPointList], stim_info: dict, cell_manager):
        super().__init__(target_points, stim_info, cell_manager)

    @staticmethod
    def parse_check_all_parameters(_stim_info: dict):
        return True

    def compute_parameters(self, cell):
        hypamp = cell.getHypAmp()
        self.amp_start = hypamp
        self.amp_end = hypamp


@StimulusManager.register_type
class RelativeLinear(Linear):
    """Injects a linear current ramp relative to cell threshold."""

    def __init__(self, target_points: list[TargetPointList], stim_info: dict, cell_manager):
        super().__init__(target_points, stim_info, cell_manager)

    def parse_check_all_parameters(self, stim_info: dict):
        # Amplitude at start as percent of threshold
        self.perc_start = float(stim_info["PercentStart"])

        # Amplitude at end as percent of threshold (optional, else same as start)
        self.perc_end = float(stim_info.get("PercentEnd", self.perc_start))

        return self.perc_start != 0 or self.perc_end != 0  # no-op if both 0

    def compute_parameters(self, cell):
        threshold = cell.getThreshold()
        # here we use parentheses to match HOC exactly
        self.amp_start = threshold * (self.perc_start / 100)
        self.amp_end = threshold * (self.perc_end / 100)


@StimulusManager.register_type
class SubThreshold(Linear):
    """Injects a current step at some percent below a cell's threshold."""

    def __init__(self, target_points: list[TargetPointList], stim_info: dict, cell_manager):
        super().__init__(target_points, stim_info, cell_manager)

    def parse_check_all_parameters(self, stim_info: dict):
        # amplitude as percent below threshold = 100%
        self.perc_less = float(stim_info["PercentLess"])

        return True

    def compute_parameters(self, cell):
        threshold = cell.getThreshold()
        # here we use parentheses to match HOC exactly
        self.amp_start = threshold * (100 - self.perc_less) / 100
        self.amp_end = self.amp_start


@StimulusManager.register_type
class Noise(BaseStim):
    """Inject a noisy (gaussian) current step, relative to cell threshold or not."""

    IsNoise = True
    stim_count = 0  # global count for seeding

    def __init__(self, target_points: list[TargetPointList], stim_info: dict, cell_manager):
        super().__init__(target_points, stim_info, cell_manager)

        self.stimList = []  # CurrentSource's go here

        self.parse_check_all_parameters(stim_info)

        sim_dt = float(SimConfig.run_conf["Dt"])  # simulation time-step [ms]

        # apply stim to each point in target_points
        for target_point_list in target_points:
            gid = target_point_list.gid
            cell = cell_manager.get_cell(gid)

            self.compute_parameters(cell)

            rng = random.Random123(
                Noise.stim_count + 100, SimConfig.rng_info.getStimulusSeed() + 500, gid + 300
            )

            # draw already used numbers
            if self.delay > 0:
                self.draw_already_used_numbers(rng, sim_dt)

            for sec_id, sc in enumerate(target_point_list.sclst):
                # skip sections not in this split
                if not sc.exists():
                    continue

                # generate noise current source
                cs = CurrentSource.noise(
                    self.mean,
                    self.var,
                    self.duration,
                    dt=self.dt,
                    delay=self.delay,
                    rng=rng,
                    represents_physical_electrode=self.represents_physical_electrode,
                )
                # attach current source to section
                cs.attach_to(sc.sec, target_point_list.x[sec_id])
                self.stimList.append(cs)  # save CurrentSource

        Noise.stim_count += 1  # increment global count

    def parse_check_all_parameters(self, stim_info: dict):
        self.dt = float(stim_info.get("Dt", 0.5))  # stimulus timestep [ms]
        if self.dt <= 0:
            raise Exception(f"{self.__class__.__name__} time-step must be positive")

        if "Mean" in stim_info:
            self.is_relative = False
            self.mean = float(stim_info["Mean"])  # noise current mean [nA]

            self.var = float(stim_info["Variance"])  # noise current variance [nA]
            if self.var <= 0:
                raise Exception(f"{self.__class__.__name__} variance must be positive")
        else:
            self.is_relative = True
            # noise current mean as percent of threshold
            self.mean_perc = float(stim_info["MeanPercent"])

            # noise current variance as percent of threshold
            self.var_perc = float(stim_info["Variance"])
            if self.var_perc <= 0:
                raise Exception(f"{self.__class__.__name__} variance percent must be positive")

        return True

    def compute_parameters(self, cell):
        if self.is_relative:
            threshold = cell.getThreshold()  # threshold current [nA]
            # here threshold MUST be first factor to match HOC exactly
            self.mean = threshold * self.mean_perc / 100
            # note that here variance has units of nA, not nA^2
            self.var = threshold * self.var_perc / 100

    def draw_already_used_numbers(self, rng, dt):
        prev_t = 0
        tstep = self.duration - dt

        while prev_t < self.delay - dt:
            next_t = min(prev_t + tstep, self.delay - dt)

            tvec = Nd.h.Vector()
            tvec.indgen(prev_t, next_t, self.dt)
            stim = Nd.h.Vector(len(tvec))
            stim.setrand(rng)

            prev_t = next_t + dt


@StimulusManager.register_type
class Pulse(BaseStim):
    """Inject a pulse train with given amplitude, frequency and width."""

    def __init__(self, target_points: list[TargetPointList], stim_info: dict, cell_manager):
        super().__init__(target_points, stim_info, cell_manager)

        self.stimList = []  # CurrentSource's go here

        if not self.parse_check_all_parameters(stim_info):
            return  # nothing to do, stim is a no-op

        # apply stim to each point in target_points
        for target_point_list in target_points:
            for sec_id, sc in enumerate(target_point_list.sclst):
                # skip sections not in this split
                if not sc.exists():
                    continue

                # generate pulse train current source
                cs = CurrentSource.train(
                    self.amp,
                    self.freq,
                    self.width,
                    self.duration,
                    delay=self.delay,
                    represents_physical_electrode=self.represents_physical_electrode,
                )
                # attach current source to section
                cs.attach_to(sc.sec, target_point_list.x[sec_id])
                self.stimList.append(cs)  # save CurrentSource

    def parse_check_all_parameters(self, stim_info: dict):
        self.amp = float(stim_info["AmpStart"])  # amplitude [nA]
        self.freq = float(stim_info["Frequency"])  # frequency [Hz]
        self.width = float(stim_info["Width"])  # pulse width [ms]

        return self.freq > 0 and self.width > 0  # no-op if any is 0


@StimulusManager.register_type
class Sinusoidal(BaseStim):
    """Inject a sinusoidal current with given amplitude and frequency."""

    def __init__(self, target_points: list[TargetPointList], stim_info: dict, cell_manager):
        super().__init__(target_points, stim_info, cell_manager)

        self.stimList = []  # CurrentSource's go here

        if not self.parse_check_all_parameters(stim_info):
            return  # nothing to do, stim is a no-op

        # apply stim to each point in target_points
        for target_point_list in target_points:
            for sec_id, sc in enumerate(target_point_list.sclst):
                # skip sections not in this split
                if not sc.exists():
                    continue

                # generate sinusoidal current source
                cs = CurrentSource.sin(
                    self.amp,
                    self.duration,
                    self.freq,
                    step=self.dt,
                    delay=self.delay,
                    represents_physical_electrode=self.represents_physical_electrode,
                )
                # attach current source to section
                cs.attach_to(sc.sec, target_point_list.x[sec_id])
                self.stimList.append(cs)  # save CurrentSource

    def parse_check_all_parameters(self, stim_info: dict):
        self.dt = float(stim_info.get("Dt", 0.025))  # stimulus timestep [ms]
        if self.dt <= 0:
            raise Exception(f"{self.__class__.__name__} time-step must be positive")

        self.amp = float(stim_info["AmpStart"])  # amplitude [nA]
        self.freq = float(stim_info["Frequency"])  # frequency [Hz]

        return self.freq > 0  # no-op if 0


@StimulusManager.register_type
class SEClamp(BaseStim):
    """Apply a single electrode voltage clamp."""

    def __init__(self, target_points: list[TargetPointList], stim_info: dict, cell_manager):
        super().__init__(target_points, stim_info, cell_manager)

        self.stimList = []  # SEClamp's go here

        self.parse_check_all_parameters(stim_info)

        # apply stim to each point in target_points
        for target_point_list in target_points:
            for sec_id, sc in enumerate(target_point_list.sclst):
                # skip sections not in this split
                if not sc.exists():
                    continue

                # create single electrode voltage clamp at location
                seclamp = Nd.h.SEClamp(target_point_list.x[sec_id], sec=sc.sec)

                seclamp.rs = self.rs
                seclamp.dur1 = self.duration
                seclamp.amp1 = self.vhold
                self.stimList.append(seclamp)  # save SEClamp

    def parse_check_all_parameters(self, stim_info: dict):
        self.vhold = float(stim_info["Voltage"])  # holding voltage [mV]
        self.rs = float(stim_info.get("RS", 0.01))  # series resistance [MOhm]
        if self.delay > 0:
            logging.warning("%s ignores delay", self.__class__.__name__)
